from abc import ABCMeta, abstractmethod


class LocalMembershipStateABC(metaclass=ABCMeta):
    @staticmethod
    @abstractmethod
    def has_members():
        raise NotImplementedError


class NoInfo(LocalMembershipStateABC):
    @staticmethod
    def has_members():
        return False


class Include(LocalMembershipStateABC):
    @staticmethod
    def has_members():
        return True


class LocalMembership():
    NoInfo = NoInfo()
    Include = Include()


from time import time

try:
    from threading import _Timer as Timer
except ImportError:
    from threading import Timer

class RemainingTimer(Timer):
    def __init__(self, interval, function):
        super().__init__(interval, function)
        self.start_time = time()

    def time_remaining(self):
        delta_time = time() - self.start_time
        return self.interval - delta_time


'''
def test():
    print("ola")

x = RemainingTimer(10, test)
x.start()
from time import sleep
for i in range(0, 10):
    print(x.time_remaining())
    sleep(1)
'''



import struct
import socket
import ipaddress
import subprocess
from ctypes import create_string_buffer, addressof

SO_ATTACH_FILTER = 26
ETH_P_IP = 0x0800    # Internet Protocol packet
ETH_P_IPV6 = 0x86DD  # IPv6 over bluebook

SO_RCVBUFFORCE = 33

def get_s_g_bpf_filter_code(source, group, interface_name):
    ip_source_version = ipaddress.ip_address(source).version
    ip_group_version = ipaddress.ip_address(source).version
    if ip_source_version == ip_group_version == 4:
        # cmd = "tcpdump -ddd \"(udp or icmp) and host %s and dst %s\"" % (source, group)
        cmd = "tcpdump -ddd \"(ip proto not 2) and host %s and dst %s\"" % (source, group)
        protocol = ETH_P_IP
    elif ip_source_version == ip_group_version == 6:
        # TODO: allow ICMPv6 echo request/echo response to be considered multicast packets
        cmd = "tcpdump -ddd \"(ip6 proto not 58) and host %s and dst %s\"" % (source, group)
        protocol = ETH_P_IPV6
    else:
        raise Exception("Unknown IP family")

    result = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
    bpf_filter = b''

    tmp = result.stdout.read().splitlines()
    num = int(tmp[0])
    for line in tmp[1:]:
        print(line)
        bpf_filter += struct.pack("HBBI", *tuple(map(int, line.split(b' '))))

    print(num)

    # defined in linux/filter.h.
    b = create_string_buffer(bpf_filter)
    mem_addr_of_filters = addressof(b)
    fprog = struct.pack('HL', num, mem_addr_of_filters)


    # Create listening socket with filters
    s = socket.socket(socket.AF_PACKET, socket.SOCK_RAW, protocol)
    s.setsockopt(socket.SOL_SOCKET, SO_ATTACH_FILTER, fprog)
    # todo pequeno ajuste (tamanho de buffer pequeno para o caso de trafego em rajadas):
    #s.setsockopt(socket.SOL_SOCKET, SO_RCVBUFFORCE, 1)
    s.bind((interface_name, protocol))

    return s



import os
import socket
import struct
import ipaddress
import traceback
from socket import if_nametoindex
from threading import RLock, Thread
from abc import abstractmethod, ABCMeta

from pimdm import UnicastRouting, Main
from pimdm.rwlock.RWLock import RWLockWrite
from pimdm.tree import pim_globals

from mld.InterfaceMLD import InterfaceMLD
from igmp.InterfaceIGMP import InterfaceIGMP
from pimdm.InterfacePIM import InterfacePim
from pimdm.InterfacePIM6 import InterfacePim6
from pimdm.tree.KernelEntry import KernelEntry
from pimdm.tree.KernelEntryInterface import KernelEntry4Interface, KernelEntry6Interface


class Kernel(metaclass=ABCMeta):
    # Max Number of Virtual Interfaces
    MAXVIFS = 32

    def __init__(self, kernel_socket):
        # Kernel is running
        self.running = True

        # KEY : interface_ip, VALUE : vif_index
        self.vif_index_to_name_dic = {}  # KEY : vif_index, VALUE : interface_name
        self.vif_name_to_index_dic = {}  # KEY : interface_name, VALUE : vif_index

        # KEY : source_ip, VALUE : {group_ip: KernelEntry}
        self.routing = {}

        self.socket = kernel_socket
        self.rwlock = RWLockWrite()
        self.interface_lock = RLock()

        # Create register interface
        # todo useless in PIM-DM... useful in PIM-SM
        #self.create_virtual_interface("0.0.0.0", "pimreg", index=0, flags=Kernel.VIFF_REGISTER)

        # interfaces being monitored by this process
        self.pim_interface = {}  # name: interface_pim
        self.membership_interface = {}  # name: interface_igmp or interface_mld

        # logs
        self.interface_logger = Main.logger.getChild('KernelInterface')
        self.tree_logger = Main.logger.getChild('KernelTree')

        # receive signals from kernel with a background thread
        handler_thread = Thread(target=self.handler)
        handler_thread.daemon = True
        handler_thread.start()

    '''
    Structure to create/remove virtual interfaces
    struct vifctl {
        vifi_t	vifc_vifi;		        /* Index of VIF */
        unsigned char vifc_flags;	    /* VIFF_ flags */
        unsigned char vifc_threshold;	/* ttl limit */
        unsigned int vifc_rate_limit;	/* Rate limiter values (NI) */
        union {
            struct in_addr vifc_lcl_addr;     /* Local interface address */
            int            vifc_lcl_ifindex;  /* Local interface index   */
        };
        struct in_addr vifc_rmt_addr;	/* IPIP tunnel addr */
    };
    '''
    @abstractmethod
    def create_virtual_interface(self, ip_interface: str or bytes, interface_name: str, index, flags=0x0):
        raise NotImplementedError

    def create_pim_interface(self, interface_name: str, state_refresh_capable:bool):
        with self.interface_lock:
            pim_interface = self.pim_interface.get(interface_name)
            membership_interface = self.membership_interface.get(interface_name)
            vif_already_exists = pim_interface or membership_interface
            if pim_interface:
                # already exists
                pim_interface.set_state_refresh_capable(state_refresh_capable)
                return
            elif membership_interface:
                index = membership_interface.vif_index
            else:
                index = list(range(0, self.MAXVIFS) - self.vif_index_to_name_dic.keys())[0]

            ip_interface = None
            if interface_name not in self.pim_interface:
                pim_interface = self._create_pim_interface_object(interface_name, index, state_refresh_capable)
                self.pim_interface[interface_name] = pim_interface
                ip_interface = pim_interface.ip_interface

            if not vif_already_exists:
                self.create_virtual_interface(ip_interface=ip_interface, interface_name=interface_name, index=index)

    @abstractmethod
    def _create_pim_interface_object(self, interface_name, index, state_refresh_capable):
        raise NotImplementedError

    def create_membership_interface(self, interface_name: str):
        with self.interface_lock:
            pim_interface = self.pim_interface.get(interface_name)
            membership_interface = self.membership_interface.get(interface_name)
            vif_already_exists = pim_interface or membership_interface
            if membership_interface:
                # already exists
                return
            elif pim_interface:
                index = pim_interface.vif_index
            else:
                index = list(range(0, self.MAXVIFS) - self.vif_index_to_name_dic.keys())[0]

            if interface_name not in self.membership_interface:
                membership_interface = self._create_membership_interface_object(interface_name, index)
                self.membership_interface[interface_name] = membership_interface
                ip_interface = membership_interface.ip_interface

                if not vif_already_exists:
                    self.create_virtual_interface(ip_interface=ip_interface, interface_name=interface_name, index=index)
                membership_interface.enable()

    @abstractmethod
    def _create_membership_interface_object(self, interface_name, index):
        raise NotImplementedError

    def remove_interface(self, interface_name, membership: bool = False, pim: bool = False):
        with self.interface_lock:
            pim_interface = self.pim_interface.get(interface_name)
            membership_interface = self.membership_interface.get(interface_name)
            if (membership and not membership_interface) or (pim and not pim_interface) or (not membership and not pim):
                return
            if pim:
                pim_interface = self.pim_interface.pop(interface_name)
                pim_interface.remove()
            elif membership:
                membership_interface = self.membership_interface.pop(interface_name)
                membership_interface.remove()

            if not self.membership_interface.get(interface_name) and not self.pim_interface.get(interface_name):
                self.remove_virtual_interface(interface_name)

    @abstractmethod
    def remove_virtual_interface(self, interface_name):
        raise NotImplementedError

    #############################################
    # Manipulate multicast routing table
    #############################################
    @abstractmethod
    def set_multicast_route(self, kernel_entry: KernelEntry):
        raise NotImplementedError

    @abstractmethod
    def set_flood_multicast_route(self, source_ip, group_ip, inbound_interface_index):
        raise NotImplementedError

    @abstractmethod
    def remove_multicast_route(self, kernel_entry: KernelEntry):
        raise NotImplementedError

    @abstractmethod
    def exit(self):
        raise NotImplementedError

    @abstractmethod
    def handler(self):
        raise NotImplementedError

    def get_routing_entry(self, source_group: tuple, create_if_not_existent=True):
        ip_src = source_group[0]
        ip_dst = source_group[1]
        with self.rwlock.genRlock():
            if ip_src in self.routing and ip_dst in self.routing[ip_src]:
                return self.routing[ip_src][ip_dst]

        with self.rwlock.genWlock():
            if ip_src in self.routing and ip_dst in self.routing[ip_src]:
                return self.routing[ip_src][ip_dst]
            elif create_if_not_existent:
                kernel_entry = KernelEntry(ip_src, ip_dst, self._get_kernel_entry_interface())
                if ip_src not in self.routing:
                    self.routing[ip_src] = {}

                iif = UnicastRouting.check_rpf(ip_src)
                self.set_flood_multicast_route(ip_src, ip_dst, iif)
                self.routing[ip_src][ip_dst] = kernel_entry
                return kernel_entry
            else:
                return None

    @staticmethod
    @abstractmethod
    def _get_kernel_entry_interface():
        pass

    # notify KernelEntries about changes at the unicast routing table
    def notify_unicast_changes(self, subnet):
        with self.rwlock.genWlock():
            for source_ip in list(self.routing.keys()):
                source_ip_obj = ipaddress.ip_address(source_ip)
                if source_ip_obj not in subnet:
                    continue
                for group_ip in list(self.routing[source_ip].keys()):
                    self.routing[source_ip][group_ip].network_update()


    # notify about changes at the interface (IP)
    '''
    def notify_interface_change(self, interface_name):
        with self.interface_lock:
            # check if interface was already added
            if interface_name not in self.vif_name_to_index_dic:
                return

            print("trying to change ip")
            pim_interface = self.pim_interface.get(interface_name)
            if pim_interface:
                old_ip = pim_interface.get_ip()
                pim_interface.change_interface()
                new_ip = pim_interface.get_ip()
                if old_ip != new_ip:
                    self.vif_dic[new_ip] = self.vif_dic.pop(old_ip)

            igmp_interface = self.igmp_interface.get(interface_name)
            if igmp_interface:
                igmp_interface.change_interface()
    '''

    # When interface changes number of neighbors verify if olist changes and prune/forward respectively
    def interface_change_number_of_neighbors(self):
        with self.rwlock.genRlock():
            for groups_dict in self.routing.values():
                for entry in groups_dict.values():
                    entry.change_at_number_of_neighbors()

    # When new neighbor connects try to resend last state refresh msg (if AssertWinner)
    def new_or_reset_neighbor(self, vif_index, neighbor_ip):
        with self.rwlock.genRlock():
            for groups_dict in self.routing.values():
                for entry in groups_dict.values():
                    entry.new_or_reset_neighbor(vif_index, neighbor_ip)


class Kernel4(Kernel):
    # MRT
    MRT_BASE    = 200
    MRT_INIT    = (MRT_BASE)      # /* Activate the kernel mroute code 	*/
    MRT_DONE    = (MRT_BASE + 1)  # /* Shutdown the kernel mroute		*/
    MRT_ADD_VIF = (MRT_BASE + 2)  # /* Add a virtual interface		    */
    MRT_DEL_VIF = (MRT_BASE + 3)  # /* Delete a virtual interface		*/
    MRT_ADD_MFC = (MRT_BASE + 4)  # /* Add a multicast forwarding entry	*/
    MRT_DEL_MFC = (MRT_BASE + 5)  # /* Delete a multicast forwarding entry	*/
    MRT_VERSION = (MRT_BASE + 6)  # /* Get the kernel multicast version	*/
    MRT_ASSERT  = (MRT_BASE + 7)  # /* Activate PIM assert mode		    */
    MRT_PIM     = (MRT_BASE + 8)  # /* enable PIM code			        */
    MRT_TABLE   = (MRT_BASE + 9)  # /* Specify mroute table ID		    */
    #MRT_ADD_MFC_PROXY = (MRT_BASE + 10)  # /* Add a (*,*|G) mfc entry	*/
    #MRT_DEL_MFC_PROXY = (MRT_BASE + 11)  # /* Del a (*,*|G) mfc entry	*/
    #MRT_MAX = (MRT_BASE + 11)

    # Max Number of Virtual Interfaces
    MAXVIFS = 32

    # SIGNAL MSG TYPE
    IGMPMSG_NOCACHE = 1
    IGMPMSG_WRONGVIF = 2
    IGMPMSG_WHOLEPKT = 3  # NOT USED ON PIM-DM

    # Interface flags
    VIFF_TUNNEL      = 0x1  # IPIP tunnel
    VIFF_SRCRT       = 0x2  # NI
    VIFF_REGISTER    = 0x4  # register vif
    VIFF_USE_IFINDEX = 0x8  # use vifc_lcl_ifindex instead of vifc_lcl_addr to find an interface

    def __init__(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_RAW, socket.IPPROTO_IGMP)

        # MRT TABLE
        if pim_globals.MULTICAST_TABLE_ID != 0:
            try:
                s.setsockopt(socket.IPPROTO_IP, self.MRT_TABLE, pim_globals.MULTICAST_TABLE_ID)
            except:
                traceback.print_exc()

        # MRT INIT
        s.setsockopt(socket.IPPROTO_IP, self.MRT_INIT, 1)

        # MRT PIM
        s.setsockopt(socket.IPPROTO_IP, self.MRT_PIM, 0)
        s.setsockopt(socket.IPPROTO_IP, self.MRT_ASSERT, 1)

        super().__init__(s)

    '''
    Structure to create/remove virtual interfaces
    struct vifctl {
        vifi_t	vifc_vifi;		        /* Index of VIF */
        unsigned char vifc_flags;	    /* VIFF_ flags */
        unsigned char vifc_threshold;	/* ttl limit */
        unsigned int vifc_rate_limit;	/* Rate limiter values (NI) */
        union {
            struct in_addr vifc_lcl_addr;     /* Local interface address */
            int            vifc_lcl_ifindex;  /* Local interface index   */
        };
        struct in_addr vifc_rmt_addr;	/* IPIP tunnel addr */
    };
    '''
    def create_virtual_interface(self, ip_interface: str or bytes, interface_name: str, index, flags=0x0):
        if type(ip_interface) is str:
            ip_interface = socket.inet_aton(ip_interface)

        struct_mrt_add_vif = struct.pack("HBBI 4s 4s", index, flags, 1, 0, ip_interface,
                                         socket.inet_aton("0.0.0.0"))
        os.system("ip mrule del iif {}".format(interface_name))
        os.system("ip mrule del oif {}".format(interface_name))
        if pim_globals.MULTICAST_TABLE_ID != 0:
            os.system("ip mrule add iif {} lookup {}".format(interface_name, pim_globals.MULTICAST_TABLE_ID))
            os.system("ip mrule add oif {} lookup {}".format(interface_name, pim_globals.MULTICAST_TABLE_ID))
        with self.rwlock.genWlock():
            self.socket.setsockopt(socket.IPPROTO_IP, self.MRT_ADD_VIF, struct_mrt_add_vif)
            self.vif_index_to_name_dic[index] = interface_name
            self.vif_name_to_index_dic[interface_name] = index

            for source_dict in list(self.routing.values()):
                for kernel_entry in list(source_dict.values()):
                    kernel_entry.new_interface(index)

        self.interface_logger.debug('Create virtual interface: %s -> %d', interface_name, index)
        return index

    def remove_virtual_interface(self, interface_name):
        #with self.interface_lock:
        index = self.vif_name_to_index_dic.pop(interface_name, None)
        struct_vifctl = struct.pack("HBBI 4s 4s", index, 0, 0, 0, socket.inet_aton("0.0.0.0"), socket.inet_aton("0.0.0.0"))

        os.system("ip mrule del iif {}".format(interface_name))
        os.system("ip mrule del oif {}".format(interface_name))
        self.socket.setsockopt(socket.IPPROTO_IP, self.MRT_DEL_VIF, struct_vifctl)

        del self.vif_name_to_index_dic[self.vif_index_to_name_dic[index]]
        interface_name = self.vif_index_to_name_dic.pop(index)

        # change MFC's to not forward traffic by this interface (set OIL to 0 for this interface)
        with self.rwlock.genWlock():
            for source_dict in list(self.routing.values()):
                for kernel_entry in list(source_dict.values()):
                    kernel_entry.remove_interface(index)

        self.interface_logger.debug('Remove virtual interface: %s -> %d', interface_name, index)



    '''
    /* Cache manipulation structures for mrouted and PIMd */
    struct mfcctl {
        struct in_addr mfcc_origin;		    /* Origin of mcast	*/
        struct in_addr mfcc_mcastgrp;		/* Group in question	*/
        vifi_t mfcc_parent; 			    /* Where it arrived	*/
        unsigned char mfcc_ttls[MAXVIFS];	/* Where it is going	*/
        unsigned int mfcc_pkt_cnt;		    /* pkt count for src-grp */
        unsigned int mfcc_byte_cnt;
        unsigned int mfcc_wrong_if;
        int mfcc_expire;
    };
    '''
    def set_multicast_route(self, kernel_entry: KernelEntry):
        source_ip = socket.inet_aton(kernel_entry.source_ip)
        group_ip = socket.inet_aton(kernel_entry.group_ip)

        outbound_interfaces = kernel_entry.get_outbound_interfaces_indexes()
        if len(outbound_interfaces) != Kernel.MAXVIFS:
            raise Exception

        #outbound_interfaces_and_other_parameters = list(kernel_entry.outbound_interfaces) + [0]*4
        outbound_interfaces_and_other_parameters = outbound_interfaces + [0]*4

        #outbound_interfaces, 0, 0, 0, 0 <- only works with python>=3.5
        #struct_mfcctl = struct.pack("4s 4s H " + "B"*Kernel.MAXVIFS + " IIIi", source_ip, group_ip, inbound_interface_index, *outbound_interfaces, 0, 0, 0, 0)
        struct_mfcctl = struct.pack("4s 4s H " + "B"*Kernel.MAXVIFS + " IIIi", source_ip, group_ip, kernel_entry.inbound_interface_index, *outbound_interfaces_and_other_parameters)
        self.socket.setsockopt(socket.IPPROTO_IP, self.MRT_ADD_MFC, struct_mfcctl)

    def set_flood_multicast_route(self, source_ip, group_ip, inbound_interface_index):
        source_ip = socket.inet_aton(source_ip)
        group_ip = socket.inet_aton(group_ip)

        outbound_interfaces = [1]*self.MAXVIFS
        outbound_interfaces[inbound_interface_index] = 0

        #outbound_interfaces_and_other_parameters = list(kernel_entry.outbound_interfaces) + [0]*4
        outbound_interfaces_and_other_parameters = outbound_interfaces + [0]*3 + [20]

        #outbound_interfaces, 0, 0, 0, 0 <- only works with python>=3.5
        #struct_mfcctl = struct.pack("4s 4s H " + "B"*Kernel.MAXVIFS + " IIIi", source_ip, group_ip, inbound_interface_index, *outbound_interfaces, 0, 0, 0, 0)
        struct_mfcctl = struct.pack("4s 4s H " + "B"*Kernel.MAXVIFS + " IIIi", source_ip, group_ip, inbound_interface_index, *outbound_interfaces_and_other_parameters)
        self.socket.setsockopt(socket.IPPROTO_IP, self.MRT_ADD_MFC, struct_mfcctl)

    def remove_multicast_route(self, kernel_entry: KernelEntry):
        source_ip = socket.inet_aton(kernel_entry.source_ip)
        group_ip = socket.inet_aton(kernel_entry.group_ip)
        outbound_interfaces_and_other_parameters = [0] + [0]*Kernel.MAXVIFS + [0]*4

        struct_mfcctl = struct.pack("4s 4s H " + "B"*Kernel.MAXVIFS + " IIIi", source_ip, group_ip, *outbound_interfaces_and_other_parameters)
        self.socket.setsockopt(socket.IPPROTO_IP, self.MRT_DEL_MFC, struct_mfcctl)
        self.routing[kernel_entry.source_ip].pop(kernel_entry.group_ip)
        if len(self.routing[kernel_entry.source_ip]) == 0:
            self.routing.pop(kernel_entry.source_ip)

    def exit(self):
        self.running = False

        # MRT DONE
        self.socket.setsockopt(socket.IPPROTO_IP, self.MRT_DONE, 1)
        self.socket.close()


    '''
    /* This is the format the mroute daemon expects to see IGMP control
    * data. Magically happens to be like an IP packet as per the original
    */
    struct igmpmsg {
        __u32 unused1,unused2;
        unsigned char im_msgtype;		/* What is this */
        unsigned char im_mbz;			/* Must be zero */
        unsigned char im_vif;			/* Interface (this ought to be a vifi_t!) */
        unsigned char unused3;
        struct in_addr im_src,im_dst;
    };
    '''
    def handler(self):
        while self.running:
            try:
                msg = self.socket.recv(20)
                (_, _, im_msgtype, im_mbz, im_vif, _, im_src, im_dst) = struct.unpack("II B B B B 4s 4s", msg[:20])
                print((im_msgtype, im_mbz, socket.inet_ntoa(im_src), socket.inet_ntoa(im_dst)))

                if im_mbz != 0:
                    continue

                print(im_msgtype)
                print(im_mbz)
                print(im_vif)
                print(socket.inet_ntoa(im_src))
                print(socket.inet_ntoa(im_dst))
                #print((im_msgtype, im_mbz, socket.inet_ntoa(im_src), socket.inet_ntoa(im_dst)))

                ip_src = socket.inet_ntoa(im_src)
                ip_dst = socket.inet_ntoa(im_dst)

                if im_msgtype == self.IGMPMSG_NOCACHE:
                    print("IGMP NO CACHE")
                    self.igmpmsg_nocache_handler(ip_src, ip_dst, im_vif)
                elif im_msgtype == self.IGMPMSG_WRONGVIF:
                    print("WRONG VIF HANDLER")
                    self.igmpmsg_wrongvif_handler(ip_src, ip_dst, im_vif)
                #elif im_msgtype == Kernel.IGMPMSG_WHOLEPKT:
                #    print("IGMP_WHOLEPKT")
                #    self.igmpmsg_wholepacket_handler(ip_src, ip_dst)
                else:
                    raise Exception
            except Exception:
                traceback.print_exc()
                continue

    # receive multicast (S,G) packet and multicast routing table has no (S,G) entry
    def igmpmsg_nocache_handler(self, ip_src, ip_dst, iif):
        source_group_pair = (ip_src, ip_dst)
        self.get_routing_entry(source_group_pair, create_if_not_existent=True).recv_data_msg(iif)

    # receive multicast (S,G) packet in a outbound_interface
    def igmpmsg_wrongvif_handler(self, ip_src, ip_dst, iif):
        source_group_pair = (ip_src, ip_dst)
        self.get_routing_entry(source_group_pair, create_if_not_existent=True).recv_data_msg(iif)


    ''' useless in PIM-DM... useful in PIM-SM
    def igmpmsg_wholepacket_handler(self, ip_src, ip_dst):
        #kernel_entry = self.routing[(ip_src, ip_dst)]
        source_group_pair = (ip_src, ip_dst)
        self.get_routing_entry(source_group_pair, create_if_not_existent=True).recv_data_msg()
        #kernel_entry.recv_data_msg(iif)
    '''

    @staticmethod
    def _get_kernel_entry_interface():
        return KernelEntry4Interface

    def _create_pim_interface_object(self, interface_name, index, state_refresh_capable):
        return InterfacePim(interface_name, index, state_refresh_capable)

    def _create_membership_interface_object(self, interface_name, index):
        return InterfaceIGMP(interface_name, index)


class Kernel6(Kernel):
    # MRT6
    MRT6_BASE = 200
    MRT6_INIT = (MRT6_BASE)                 # /* Activate the kernel mroute code 	*/
    MRT6_DONE = (MRT6_BASE + 1)             # /* Shutdown the kernel mroute		*/
    MRT6_ADD_MIF = (MRT6_BASE + 2)          # /* Add a virtual interface		*/
    MRT6_DEL_MIF = (MRT6_BASE + 3)          # /* Delete a virtual interface		*/
    MRT6_ADD_MFC = (MRT6_BASE + 4)          # /* Add a multicast forwarding entry	*/
    MRT6_DEL_MFC = (MRT6_BASE + 5)          # /* Delete a multicast forwarding entry	*/
    MRT6_VERSION = (MRT6_BASE + 6)          # /* Get the kernel multicast version	*/
    MRT6_ASSERT = (MRT6_BASE + 7)           # /* Activate PIM assert mode		*/
    MRT6_PIM = (MRT6_BASE + 8)              # /* enable PIM code			*/
    MRT6_TABLE = (MRT6_BASE + 9)            # /* Specify mroute table ID		*/
    MRT6_ADD_MFC_PROXY = (MRT6_BASE + 10)   # /* Add a (*,*|G) mfc entry	*/
    MRT6_DEL_MFC_PROXY = (MRT6_BASE + 11)   # /* Del a (*,*|G) mfc entry	*/
    MRT6_MAX = (MRT6_BASE + 11)

    # define SIOCGETMIFCNT_IN6	SIOCPROTOPRIVATE	/* IP protocol privates */
    # define SIOCGETSGCNT_IN6	(SIOCPROTOPRIVATE+1)
    # define SIOCGETRPF	(SIOCPROTOPRIVATE+2)

    # Max Number of Virtual Interfaces
    MAXVIFS = 32

    # SIGNAL MSG TYPE
    MRT6MSG_NOCACHE = 1
    MRT6MSG_WRONGMIF = 2
    MRT6MSG_WHOLEPKT = 3  # /* used for use level encap */

    # Interface flags
    MIFF_REGISTER = 0x1  # /* register vif	*/

    def __init__(self):
        s = socket.socket(socket.AF_INET6, socket.SOCK_RAW, socket.IPPROTO_ICMPV6)

        # MRT TABLE
        if pim_globals.MULTICAST_TABLE_ID != 0:
            try:
                s.setsockopt(socket.IPPROTO_IPV6, self.MRT6_TABLE, pim_globals.MULTICAST_TABLE_ID)
            except:
                traceback.print_exc()

        # MRT INIT
        s.setsockopt(socket.IPPROTO_IPV6, self.MRT6_INIT, 1)

        # MRT PIM
        s.setsockopt(socket.IPPROTO_IPV6, self.MRT6_PIM, 0)
        s.setsockopt(socket.IPPROTO_IPV6, self.MRT6_ASSERT, 1)

        super().__init__(s)

    '''
    Structure to create/remove multicast interfaces
    struct mif6ctl {
        mifi_t	mif6c_mifi;		        /* Index of MIF */
        unsigned char mif6c_flags;	    /* MIFF_ flags */
        unsigned char vifc_threshold;	/* ttl limit */
        __u16	 mif6c_pifi;		    /* the index of the physical IF */
        unsigned int vifc_rate_limit;	/* Rate limiter values (NI) */
    };
    '''
    def create_virtual_interface(self, ip_interface, interface_name: str, index, flags=0x0):
        physical_if_index = if_nametoindex(interface_name)
        struct_mrt_add_vif = struct.pack("HBBHI", index, flags, 1, physical_if_index, 0)

        os.system("ip -6 mrule del iif {}".format(interface_name))
        os.system("ip -6 mrule del oif {}".format(interface_name))
        if pim_globals.MULTICAST_TABLE_ID != 0:
            os.system("ip -6 mrule add iif {} lookup {}".format(interface_name, pim_globals.MULTICAST_TABLE_ID))
            os.system("ip -6 mrule add oif {} lookup {}".format(interface_name, pim_globals.MULTICAST_TABLE_ID))

        with self.rwlock.genWlock():
            self.socket.setsockopt(socket.IPPROTO_IPV6, self.MRT6_ADD_MIF, struct_mrt_add_vif)
            self.vif_index_to_name_dic[index] = interface_name
            self.vif_name_to_index_dic[interface_name] = index

            for source_dict in list(self.routing.values()):
                for kernel_entry in list(source_dict.values()):
                    kernel_entry.new_interface(index)

        self.interface_logger.debug('Create virtual interface: %s -> %d', interface_name, index)
        return index

    def remove_virtual_interface(self, interface_name):
        # with self.interface_lock:
        mif_index = self.vif_name_to_index_dic.pop(interface_name, None)
        interface_name = self.vif_index_to_name_dic.pop(mif_index)

        physical_if_index = if_nametoindex(interface_name)
        struct_vifctl = struct.pack("HBBHI", mif_index, 0, 0, physical_if_index, 0)
        self.socket.setsockopt(socket.IPPROTO_IPV6, self.MRT6_DEL_MIF, struct_vifctl)

        os.system("ip -6 mrule del iif {}".format(interface_name))
        os.system("ip -6 mrule del oif {}".format(interface_name))

        # alterar MFC's para colocar a 0 esta interface
        with self.rwlock.genWlock():
            for source_dict in list(self.routing.values()):
                for kernel_entry in list(source_dict.values()):
                    kernel_entry.remove_interface(mif_index)

        self.interface_logger.debug('Remove virtual interface: %s -> %d', interface_name, mif_index)

    '''
    /* Cache manipulation structures for mrouted and PIMd */
    typedef	__u32 if_mask;
    typedef struct if_set {
        if_mask ifs_bits[__KERNEL_DIV_ROUND_UP(IF_SETSIZE, NIFBITS)];
    } if_set;

    struct mf6cctl {
        struct sockaddr_in6 mf6cc_origin;		/* Origin of mcast	*/
        struct sockaddr_in6 mf6cc_mcastgrp;		/* Group in question	*/
        mifi_t	mf6cc_parent;			        /* Where it arrived	*/
        struct if_set mf6cc_ifset;		        /* Where it is going */
    };
    '''
    def set_multicast_route(self, kernel_entry: KernelEntry):
        source_ip = socket.inet_pton(socket.AF_INET6, kernel_entry.source_ip)
        sockaddr_in6_source = struct.pack("H H I 16s I", socket.AF_INET6, 0, 0, source_ip, 0)
        group_ip = socket.inet_pton(socket.AF_INET6, kernel_entry.group_ip)
        sockaddr_in6_group = struct.pack("H H I 16s I", socket.AF_INET6, 0, 0, group_ip, 0)

        outbound_interfaces = kernel_entry.get_outbound_interfaces_indexes()
        if len(outbound_interfaces) != 8:
            raise Exception

        # outbound_interfaces_and_other_parameters = list(kernel_entry.outbound_interfaces) + [0]*4
        # outbound_interfaces_and_other_parameters = outbound_interfaces + [0]*4
        outgoing_interface_list = outbound_interfaces

        # outbound_interfaces, 0, 0, 0, 0 <- only works with python>=3.5
        # struct_mfcctl = struct.pack("4s 4s H " + "B"*Kernel.MAXVIFS + " IIIi", source_ip, group_ip, inbound_interface_index, *outbound_interfaces, 0, 0, 0, 0)
        struct_mf6cctl = struct.pack("28s 28s H " + "I" * 8, sockaddr_in6_source, sockaddr_in6_group,
                                     kernel_entry.inbound_interface_index,
                                     *outgoing_interface_list)
        self.socket.setsockopt(socket.IPPROTO_IPV6, self.MRT6_ADD_MFC, struct_mf6cctl)

    def set_flood_multicast_route(self, source_ip, group_ip, inbound_interface_index):
        source_ip = socket.inet_pton(socket.AF_INET6, source_ip)
        sockaddr_in6_source = struct.pack("H H I 16s I", socket.AF_INET6, 0, 0, source_ip, 0)
        group_ip = socket.inet_pton(socket.AF_INET6, group_ip)
        sockaddr_in6_group = struct.pack("H H I 16s I", socket.AF_INET6, 0, 0, group_ip, 0)

        outbound_interfaces = [255] * 8
        outbound_interfaces[inbound_interface_index // 32] = 0xFFFFFFFF & ~(1 << (inbound_interface_index % 32))

        # outbound_interfaces, 0, 0, 0, 0 <- only works with python>=3.5
        # struct_mfcctl = struct.pack("4s 4s H " + "B"*Kernel.MAXVIFS + " IIIi", source_ip, group_ip, inbound_interface_index, *outbound_interfaces, 0, 0, 0, 0)
        # struct_mfcctl = struct.pack("4s 4s H " + "B"*Kernel.MAXVIFS + " IIIi", source_ip, group_ip, inbound_interface_index, *outbound_interfaces_and_other_parameters)
        struct_mf6cctl = struct.pack("28s 28s H " + "I" * 8, sockaddr_in6_source, sockaddr_in6_group,
                                     inbound_interface_index, *outbound_interfaces)
        self.socket.setsockopt(socket.IPPROTO_IPV6, self.MRT6_ADD_MFC, struct_mf6cctl)

    def remove_multicast_route(self, kernel_entry: KernelEntry):
        source_ip = socket.inet_pton(socket.AF_INET6, kernel_entry.source_ip)
        sockaddr_in6_source = struct.pack("H H I 16s I", socket.AF_INET6, 0, 0, source_ip, 0)
        group_ip = socket.inet_pton(socket.AF_INET6, kernel_entry.group_ip)
        sockaddr_in6_group = struct.pack("H H I 16s I", socket.AF_INET6, 0, 0, group_ip, 0)
        outbound_interfaces = [0] * 8

        # struct_mfcctl = struct.pack("4s 4s H " + "B"*Kernel.MAXVIFS + " IIIi", source_ip, group_ip, *outbound_interfaces_and_other_parameters)
        struct_mf6cctl = struct.pack("28s 28s H " + "I" * 8, sockaddr_in6_source, sockaddr_in6_group, 0,
                                     *outbound_interfaces)

        self.socket.setsockopt(socket.IPPROTO_IPV6, self.MRT6_DEL_MFC, struct_mf6cctl)
        self.routing[kernel_entry.source_ip].pop(kernel_entry.group_ip)
        if len(self.routing[kernel_entry.source_ip]) == 0:
            self.routing.pop(kernel_entry.source_ip)

    def exit(self):
        self.running = False

        # MRT DONE
        self.socket.setsockopt(socket.IPPROTO_IPV6, self.MRT6_DONE, 1)
        self.socket.close()

    '''
    /*
     * Structure used to communicate from kernel to multicast router.
     * We'll overlay the structure onto an MLD header (not an IPv6 heder like igmpmsg{}
     * used for IPv4 implementation). This is because this structure will be passed via an
     * IPv6 raw socket, on which an application will only receiver the payload i.e the data after
     * the IPv6 header and all the extension headers. (See section 3 of RFC 3542)
     */

    struct mrt6msg {
        __u8		im6_mbz;		/* must be zero		   */
        __u8		im6_msgtype;	/* what type of message    */
        __u16		im6_mif;		/* mif rec'd on		   */
        __u32		im6_pad;		/* padding for 64 bit arch */
        struct in6_addr	im6_src, im6_dst;
    };

    /* ip6mr netlink cache report attributes */
    enum {
        IP6MRA_CREPORT_UNSPEC,
        IP6MRA_CREPORT_MSGTYPE,
        IP6MRA_CREPORT_MIF_ID,
        IP6MRA_CREPORT_SRC_ADDR,
        IP6MRA_CREPORT_DST_ADDR,
        IP6MRA_CREPORT_PKT,
        __IP6MRA_CREPORT_MAX
    };
    '''
    def handler(self):
        while self.running:
            try:
                msg = self.socket.recv(500)
                if len(msg) < 40:
                    continue
                (im6_mbz, im6_msgtype, im6_mif, _, im6_src, im6_dst) = struct.unpack("B B H I 16s 16s", msg[:40])
                # print((im_msgtype, im_mbz, socket.inet_ntoa(im_src), socket.inet_ntoa(im_dst)))

                if im6_mbz != 0:
                    continue

                print(im6_mbz)
                print(im6_msgtype)
                print(im6_mif)
                print(socket.inet_ntop(socket.AF_INET6, im6_src))
                print(socket.inet_ntop(socket.AF_INET6, im6_dst))
                # print((im_msgtype, im_mbz, socket.inet_ntoa(im_src), socket.inet_ntoa(im_dst)))

                ip_src = socket.inet_ntop(socket.AF_INET6, im6_src)
                ip_dst = socket.inet_ntop(socket.AF_INET6, im6_dst)

                if im6_msgtype == self.MRT6MSG_NOCACHE:
                    print("MRT6 NO CACHE")
                    self.msg_nocache_handler(ip_src, ip_dst, im6_mif)
                elif im6_msgtype == self.MRT6MSG_WRONGMIF:
                    print("WRONG MIF HANDLER")
                    self.msg_wrongvif_handler(ip_src, ip_dst, im6_mif)
                # elif im_msgtype == Kernel.IGMPMSG_WHOLEPKT:
                #    print("IGMP_WHOLEPKT")
                #    self.igmpmsg_wholepacket_handler(ip_src, ip_dst)
                else:
                    raise Exception
            except Exception:
                traceback.print_exc()
                continue

    # receive multicast (S,G) packet and multicast routing table has no (S,G) entry
    def msg_nocache_handler(self, ip_src, ip_dst, iif):
        source_group_pair = (ip_src, ip_dst)
        self.get_routing_entry(source_group_pair, create_if_not_existent=True).recv_data_msg(iif)

    # receive multicast (S,G) packet in a outbound_interface
    def msg_wrongvif_handler(self, ip_src, ip_dst, iif):
        source_group_pair = (ip_src, ip_dst)
        self.get_routing_entry(source_group_pair, create_if_not_existent=True).recv_data_msg(iif)

    ''' useless in PIM-DM... useful in PIM-SM
    def msg_wholepacket_handler(self, ip_src, ip_dst):
        #kernel_entry = self.routing[(ip_src, ip_dst)]
        source_group_pair = (ip_src, ip_dst)
        self.get_routing_entry(source_group_pair, create_if_not_existent=True).recv_data_msg()
        #kernel_entry.recv_data_msg(iif)
    '''

    @staticmethod
    def _get_kernel_entry_interface():
        return KernelEntry6Interface

    def _create_pim_interface_object(self, interface_name, index, state_refresh_capable):
        return InterfacePim6(interface_name, index, state_refresh_capable)

    def _create_membership_interface_object(self, interface_name, index):
        return InterfaceMLD(interface_name, index)



import yaml
from pimdm.tree import pim_globals
from igmp.igmp2 import igmp_globals
from mld.mld1 import mld_globals
from pimdm import Main


def parse_config_file(file_path):
    """
    Parse yaml file and set everything on protocol process accordingly
    """
    with open(file_path) as f:
        data = yaml.load(f, Loader=yaml.FullLoader)
        print(data)

        print(type(data.get("UnicastVRF", 254)))

        multicast_vrf = data.get("MulticastVRF", 0)
        pim_globals.MULTICAST_TABLE_ID = multicast_vrf
        pim_globals.UNICAST_TABLE_ID = data.get("UnicastVRF", 254)
        pim_config = data.get("PIM-DM", {})
        igmp_config = data.get("IGMP", {})
        mld_config = data.get("MLD", {})

        ##### PIM config ######
        if "DefaultTimers" in pim_config:
            pim_globals.ASSERT_TIME = pim_config["DefaultTimers"].get("ASSERT_TIME", pim_globals.ASSERT_TIME)
            pim_globals.GRAFT_RETRY_PERIOD = pim_config["DefaultTimers"].get("GRAFT_RETRY_PERIOD", pim_globals.GRAFT_RETRY_PERIOD)
            pim_globals.JP_OVERRIDE_INTERVAL = pim_config["DefaultTimers"].get("JP_OVERRIDE_INTERVAL", pim_globals.JP_OVERRIDE_INTERVAL)
            pim_globals.OVERRIDE_INTERVAL = pim_config["DefaultTimers"].get("OVERRIDE_INTERVAL", pim_globals.OVERRIDE_INTERVAL)
            pim_globals.PROPAGATION_DELAY = pim_config["DefaultTimers"].get("PROPAGATION_DELAY", pim_globals.PROPAGATION_DELAY)
            pim_globals.REFRESH_INTERVAL = pim_config["DefaultTimers"].get("REFRESH_INTERVAL", pim_globals.REFRESH_INTERVAL)
            pim_globals.SOURCE_LIFETIME = pim_config["DefaultTimers"].get("SOURCE_LIFETIME", pim_globals.SOURCE_LIFETIME)
            pim_globals.T_LIMIT = pim_config["DefaultTimers"].get("T_LIMIT", pim_globals.T_LIMIT)

        if "Interfaces" in pim_config:
            interfaces = pim_config["Interfaces"]  # type: dict

            for if_name, if_value in interfaces.items():
                if if_value.get("ipv4", False):
                    if_ipv4 = if_value["ipv4"]
                    if if_ipv4.get("enabled", False):
                        Main.add_pim_interface(interface_name=if_name, ipv4=True, ipv6=False)

                if if_value.get("ipv6", False):
                    if_ipv6 = if_value["ipv6"]
                    if if_ipv6.get("enabled", False):
                        Main.add_pim_interface(interface_name=if_name, ipv4=False, ipv6=True)


        ##### IGMP config #######
        if "Settings" in igmp_config:
            igmp_globals.ROBUSTNESS_VARIABLE = igmp_config["Settings"].get("ROBUSTNESS_VARIABLE", igmp_globals.ROBUSTNESS_VARIABLE)
            igmp_globals.QUERY_INTERVAL = igmp_config["Settings"].get("QUERY_INTERVAL", igmp_globals.QUERY_INTERVAL)
            igmp_globals.QUERY_RESPONSE_INTERVAL = igmp_config["Settings"].get("QUERY_RESPONSE_INTERVAL", igmp_globals.QUERY_RESPONSE_INTERVAL)
            igmp_globals.MAX_RESPONSE_TIME_QUERY_RESPONSE_INTERVAL = igmp_config["Settings"].get("MAX_RESPONSE_TIME_QUERY_RESPONSE_INTERVAL", igmp_globals.QUERY_RESPONSE_INTERVAL*10)
            igmp_globals.GROUP_MEMBERSHIP_INTERVAL = igmp_config["Settings"].get("GROUP_MEMBERSHIP_INTERVAL", igmp_globals.ROBUSTNESS_VARIABLE * igmp_globals.QUERY_INTERVAL + igmp_globals.QUERY_RESPONSE_INTERVAL)
            igmp_globals.OTHER_QUERIER_PRESENT_INTERVAL = igmp_config["Settings"].get("OTHER_QUERIER_PRESENT_INTERVAL", igmp_globals.ROBUSTNESS_VARIABLE * igmp_globals.QUERY_INTERVAL + igmp_globals.QUERY_RESPONSE_INTERVAL / 2)
            igmp_globals.STARTUP_QUERY_INTERVAL = igmp_config["Settings"].get("STARTUP_QUERY_INTERVAL", igmp_globals.QUERY_INTERVAL / 4)
            igmp_globals.STARTUP_QUERY_COUNT = igmp_config["Settings"].get("STARTUP_QUERY_COUNT", igmp_globals.ROBUSTNESS_VARIABLE)
            igmp_globals.LAST_MEMBER_QUERY_INTERVAL = igmp_config["Settings"].get("LAST_MEMBER_QUERY_INTERVAL", igmp_globals.LAST_MEMBER_QUERY_INTERVAL)
            igmp_globals.MAX_RESPONSE_TIME_LAST_MEMBER_QUERY_INTERVAL = igmp_config["Settings"].get("LAST_MEMBER_QUERY_COUNT", igmp_globals.LAST_MEMBER_QUERY_INTERVAL * 10)
            igmp_globals.LAST_MEMBER_QUERY_COUNT = igmp_config["Settings"].get("LAST_MEMBER_QUERY_COUNT", igmp_globals.ROBUSTNESS_VARIABLE)
            igmp_globals.UNSOLICITED_REPORT_INTERVAL = igmp_config["Settings"].get("UNSOLICITED_REPORT_INTERVAL", igmp_globals.UNSOLICITED_REPORT_INTERVAL)
            igmp_globals.VERSION_1_ROUTER_PRESENT_TIMEOUT = igmp_config["Settings"].get("VERSION_1_ROUTER_PRESENT_TIMEOUT", igmp_globals.VERSION_1_ROUTER_PRESENT_TIMEOUT)

        if "Interfaces" in igmp_config:
            interfaces = igmp_config["Interfaces"]  # type: dict

            for if_name, if_value in interfaces.items():
                if if_value.get("enabled", False):
                    Main.add_membership_interface(interface_name=if_name, ipv4=True, ipv6=False)

        ##### MLD config #######
        if "Settings" in mld_config:
            mld_globals.ROBUSTNESS_VARIABLE = mld_config["Settings"].get("ROBUSTNESS_VARIABLE", mld_globals.ROBUSTNESS_VARIABLE)
            mld_globals.QUERY_INTERVAL = mld_config["Settings"].get("QUERY_INTERVAL", mld_globals.QUERY_INTERVAL)
            mld_globals.QUERY_RESPONSE_INTERVAL = mld_config["Settings"].get("QUERY_RESPONSE_INTERVAL", mld_globals.QUERY_RESPONSE_INTERVAL)
            mld_globals.MULTICAST_LISTENER_INTERVAL = mld_config["Settings"].get("MULTICAST_LISTENER_INTERVAL", (mld_globals.ROBUSTNESS_VARIABLE * mld_globals.QUERY_INTERVAL) + (mld_globals.QUERY_RESPONSE_INTERVAL))
            mld_globals.OTHER_QUERIER_PRESENT_INTERVAL = mld_config["Settings"].get("OTHER_QUERIER_PRESENT_INTERVAL", (mld_globals.ROBUSTNESS_VARIABLE * mld_globals.QUERY_INTERVAL) + 0.5 * mld_globals.QUERY_RESPONSE_INTERVAL)
            mld_globals.STARTUP_QUERY_INTERVAL = mld_config["Settings"].get("STARTUP_QUERY_INTERVAL", (1 / 4) * mld_globals.QUERY_INTERVAL)
            mld_globals.STARTUP_QUERY_COUNT = mld_config["Settings"].get("STARTUP_QUERY_COUNT", mld_globals.ROBUSTNESS_VARIABLE)
            mld_globals.LAST_LISTENER_QUERY_INTERVAL = mld_config["Settings"].get("LAST_LISTENER_QUERY_INTERVAL", mld_globals.LAST_LISTENER_QUERY_INTERVAL)
            mld_globals.LAST_LISTENER_QUERY_COUNT = mld_config["Settings"].get("LAST_LISTENER_QUERY_COUNT", mld_globals.ROBUSTNESS_VARIABLE)
            mld_globals.UNSOLICITED_REPORT_INTERVAL = mld_config["Settings"].get("UNSOLICITED_REPORT_INTERVAL", mld_globals.UNSOLICITED_REPORT_INTERVAL)

        if "Interfaces" in mld_config:
            interfaces = mld_config["Interfaces"]  # type: dict

            for if_name, if_value in interfaces.items():
                if if_value.get("enabled", False):
                    Main.add_membership_interface(interface_name=if_name, ipv4=False, ipv6=True)


def get_yaml_file():
    """
    Get configuration file from live protocol process
    """
    dict_file = {
        'MulticastVRF': pim_globals.MULTICAST_TABLE_ID,
        'UnicastVRF': pim_globals.UNICAST_TABLE_ID,
        'PIM-DM': {
            "DefaultTimers": {
                "ASSERT_TIME": pim_globals.ASSERT_TIME,
                "GRAFT_RETRY_PERIOD": pim_globals.GRAFT_RETRY_PERIOD,
                "JP_OVERRIDE_INTERVAL": pim_globals.JP_OVERRIDE_INTERVAL,
                "OVERRIDE_INTERVAL": pim_globals.OVERRIDE_INTERVAL,
                "PROPAGATION_DELAY": pim_globals.PROPAGATION_DELAY,
                "REFRESH_INTERVAL": pim_globals.REFRESH_INTERVAL,
                "SOURCE_LIFETIME": pim_globals.SOURCE_LIFETIME,
                "T_LIMIT": pim_globals.T_LIMIT,
            },
            "Interfaces": {},
        },
        'IGMP': {
            "Settings": {
                "ROBUSTNESS_VARIABLE": igmp_globals.ROBUSTNESS_VARIABLE,
                "QUERY_INTERVAL": igmp_globals.QUERY_INTERVAL,
                "QUERY_RESPONSE_INTERVAL": igmp_globals.QUERY_RESPONSE_INTERVAL,
                "MAX_RESPONSE_TIME_QUERY_RESPONSE_INTERVAL": igmp_globals.MAX_RESPONSE_TIME_QUERY_RESPONSE_INTERVAL,
                "GROUP_MEMBERSHIP_INTERVAL": igmp_globals.GROUP_MEMBERSHIP_INTERVAL,
                "OTHER_QUERIER_PRESENT_INTERVAL": igmp_globals.OTHER_QUERIER_PRESENT_INTERVAL,
                "STARTUP_QUERY_INTERVAL": igmp_globals.STARTUP_QUERY_INTERVAL,
                "STARTUP_QUERY_COUNT": igmp_globals.STARTUP_QUERY_COUNT,
                "LAST_MEMBER_QUERY_INTERVAL": igmp_globals.LAST_MEMBER_QUERY_INTERVAL,
                "MAX_RESPONSE_TIME_LAST_MEMBER_QUERY_INTERVAL": igmp_globals.MAX_RESPONSE_TIME_LAST_MEMBER_QUERY_INTERVAL,
                "LAST_MEMBER_QUERY_COUNT": igmp_globals.LAST_MEMBER_QUERY_COUNT,
                "UNSOLICITED_REPORT_INTERVAL": igmp_globals.UNSOLICITED_REPORT_INTERVAL,
                "VERSION_1_ROUTER_PRESENT_TIMEOUT": igmp_globals.VERSION_1_ROUTER_PRESENT_TIMEOUT,
            },
            "Interfaces": {},
        },
        'MLD': {
            "Settings": {
                "ROBUSTNESS_VARIABLE": mld_globals.ROBUSTNESS_VARIABLE,
                "QUERY_INTERVAL": mld_globals.QUERY_INTERVAL,
                "QUERY_RESPONSE_INTERVAL": mld_globals.QUERY_RESPONSE_INTERVAL,
                "MULTICAST_LISTENER_INTERVAL": mld_globals.MULTICAST_LISTENER_INTERVAL,
                "OTHER_QUERIER_PRESENT_INTERVAL": mld_globals.OTHER_QUERIER_PRESENT_INTERVAL,
                "STARTUP_QUERY_INTERVAL": mld_globals.STARTUP_QUERY_INTERVAL,
                "STARTUP_QUERY_COUNT": mld_globals.STARTUP_QUERY_COUNT,
                "LAST_LISTENER_QUERY_INTERVAL": mld_globals.LAST_LISTENER_QUERY_INTERVAL,
                "LAST_LISTENER_QUERY_COUNT": mld_globals.LAST_LISTENER_QUERY_COUNT,
                "UNSOLICITED_REPORT_INTERVAL": mld_globals.UNSOLICITED_REPORT_INTERVAL,
            },
            "Interfaces": {},
        }
    }

    for if_name, if_value in Main.interfaces.items():
        dict_file["PIM-DM"]["Interfaces"][if_name] = {}
        dict_file["PIM-DM"]["Interfaces"][if_name]["ipv4"] = {
            "enabled": True,
        }

    for if_name, if_value in Main.interfaces_v6.items():
        if if_name not in dict_file["PIM-DM"]["Interfaces"]:
            dict_file["PIM-DM"]["Interfaces"][if_name] = {}

        dict_file["PIM-DM"]["Interfaces"][if_name]["ipv6"] = {
            "enabled": True,
        }

    for if_name in Main.igmp_interfaces.keys():
        dict_file["IGMP"]["Interfaces"][if_name] = {
            "enabled": True,
        }

    for if_name in Main.mld_interfaces.keys():
        dict_file["MLD"]["Interfaces"][if_name] = {
            "enabled": True,
        }

    return yaml.dump(dict_file)


def get_vrfs(file_path):
    """
    Get vrf configuration from yaml file.
    This is only used by Run.py to create the correct daemons accordingly (daemons are bound to specific VRFs).
    """
    with open(file_path) as f:
        data = yaml.load(f, Loader=yaml.FullLoader)
        multicast_vrf = data.get("MulticastVRF", 0)
        unicast_vrf = data.get("UnicastVRF", 254)
        return [multicast_vrf, unicast_vrf]



from abc import ABCMeta, abstractmethod
from threading import RLock
import traceback

from .downstream_prune import DownstreamState
from .assert_state import AssertState, AssertStateABC

from pimdm.packet.PacketPimGraft import PacketPimGraft
from pimdm.packet.PacketPimGraftAck import PacketPimGraftAck
from pimdm.packet.PacketPimJoinPruneMulticastGroup import PacketPimJoinPruneMulticastGroup
from pimdm.packet.PacketPimHeader import PacketPimHeader
from pimdm.packet.Packet import Packet

from pimdm.packet.PacketPimJoinPrune import PacketPimJoinPrune
from pimdm.packet.PacketPimAssert import PacketPimAssert
from pimdm.packet.PacketPimStateRefresh import PacketPimStateRefresh
from .metric import AssertMetric
from threading import Timer
from .local_membership import LocalMembership
from .pim_globals import T_LIMIT
import logging


class TreeInterface(metaclass=ABCMeta):
    def __init__(self, kernel_entry, interface_id, logger: logging.LoggerAdapter):
        self._kernel_entry = kernel_entry
        self._interface_id = interface_id
        self.logger = logger
        self.assert_logger = logging.LoggerAdapter(logger.logger.getChild('Assert'), logger.extra)
        self.join_prune_logger = logging.LoggerAdapter(logger.logger.getChild('JoinPrune'), logger.extra)

        # Local Membership State
        try:
            membership_interface = self.get_membership_interface()
            group_state = membership_interface.interface_state.get_group_state(kernel_entry.group_ip)
            #self._igmp_has_members = group_state.add_multicast_routing_entry(self)
            igmp_has_members = group_state.add_multicast_routing_entry(self)
            self._local_membership_state = LocalMembership.Include if igmp_has_members else LocalMembership.NoInfo
        except:
            self._local_membership_state = LocalMembership.NoInfo


        # Prune State
        self._prune_state = DownstreamState.NoInfo
        self._prune_pending_timer = None
        self._prune_timer = None

        # Assert Winner State
        self._assert_state = AssertState.NoInfo
        self._assert_winner_metric = AssertMetric()
        self._assert_timer = None
        self.assert_logger.debug("Assert state transitions to NoInfo")

        # Received prune hold time
        self._received_prune_holdtime = None

        self._membership_lock = RLock()

    ############################################
    # Set ASSERT State
    ############################################
    def set_assert_state(self, new_state: AssertStateABC):
        with self.get_state_lock():
            if new_state != self._assert_state:
                self._assert_state = new_state
                self.assert_logger.debug('Assert state transitions to ' + str(new_state))

                self.change_tree()
                self.evaluate_ingroup()

    def set_assert_winner_metric(self, new_assert_metric: AssertMetric):
        with self.get_state_lock():
            try:
                old_neighbor = self.get_interface().get_neighbor(self._assert_winner_metric.get_ip())
                new_neighbor = self.get_interface().get_neighbor(new_assert_metric.get_ip())

                if old_neighbor is not None:
                    old_neighbor.unsubscribe_nlt_expiration(self)
                if new_neighbor is not None:
                    new_neighbor.subscribe_nlt_expiration(self)
            except:
                traceback.print_exc()
            finally:
                self._assert_winner_metric = new_assert_metric

    ############################################
    # ASSERT Timer
    ############################################
    def set_assert_timer(self, time):
        self.clear_assert_timer()
        self._assert_timer = Timer(time, self.assert_timeout)
        self._assert_timer.start()

    def clear_assert_timer(self):
        if self._assert_timer is not None:
            self._assert_timer.cancel()

    def assert_timeout(self):
        self._assert_state.assertTimerExpires(self)

    ###########################################
    # Recv packets
    ###########################################
    def recv_data_msg(self):
        pass

    def recv_assert_msg(self, received_metric: AssertMetric):
        if self._assert_winner_metric.is_better_than(received_metric) and \
                self._assert_winner_metric.ip_address == received_metric.ip_address:
            # received inferior assert from Assert Winner
            self._assert_state.receivedInferiorMetricFromWinner(self)
        elif self.my_assert_metric().is_better_than(received_metric) and self.could_assert():
            # received inferior assert from non assert winner and could_assert
            self._assert_state.receivedInferiorMetricFromNonWinner_couldAssertIsTrue(self)
        elif self._assert_state.is_preferred_assert(self, received_metric):
            #received preferred assert
            equal_metric = received_metric.equal_metric(self._assert_winner_metric)
            self._assert_state.receivedPreferedMetric(self, received_metric, equal_metric)

    def recv_prune_msg(self, upstream_neighbor_address, holdtime):
        if upstream_neighbor_address == self.get_ip():
            self._assert_state.receivedPruneOrJoinOrGraft(self)

    def recv_join_msg(self, upstream_neighbor_address):
        if upstream_neighbor_address == self.get_ip():
            self._assert_state.receivedPruneOrJoinOrGraft(self)

    def recv_graft_msg(self, upstream_neighbor_address, source_ip):
        if upstream_neighbor_address == self.get_ip():
            self._assert_state.receivedPruneOrJoinOrGraft(self)

    def recv_graft_ack_msg(self, source_ip_of_graft_ack):
        return

    def recv_state_refresh_msg(self, received_metric: AssertMetric, prune_indicator):
        self.recv_assert_msg(received_metric)

    ######################################
    # Send messages
    ######################################
    def send_graft(self):
        print("send graft")
        try:
            (source, group) = self.get_tree_id()

            ip_dst = self.get_neighbor_RPF()
            ph = PacketPimGraft(ip_dst)
            ph.add_multicast_group(PacketPimJoinPruneMulticastGroup(group,  joined_src_addresses=[source]))
            pckt = Packet(payload=PacketPimHeader(ph))
            self.get_interface().send(pckt.bytes(), ip_dst)
        except:
            traceback.print_exc()
            return

    def send_graft_ack(self, ip_sender):
        print("send graft ack")
        try:
            (source, group) = self.get_tree_id()

            ph = PacketPimGraftAck(ip_sender)
            ph.add_multicast_group(PacketPimJoinPruneMulticastGroup(group,  joined_src_addresses=[source]))
            pckt = Packet(payload=PacketPimHeader(ph))
            self.get_interface().send(pckt.bytes(), ip_sender)
        except:
            traceback.print_exc()
            return

    def send_prune(self, holdtime=None):
        if holdtime is None:
            holdtime = T_LIMIT

        print("send prune")
        try:
            (source, group) = self.get_tree_id()
            ph = PacketPimJoinPrune(self.get_neighbor_RPF(), holdtime)
            ph.add_multicast_group(PacketPimJoinPruneMulticastGroup(group, pruned_src_addresses=[source]))
            pckt = Packet(payload=PacketPimHeader(ph))

            self.get_interface().send(pckt.bytes())
            print('sent prune msg')
        except:
            traceback.print_exc()
            return

    def send_pruneecho(self):
        holdtime = T_LIMIT
        try:
            (source, group) = self.get_tree_id()
            ph = PacketPimJoinPrune(self.get_ip(), holdtime)
            ph.add_multicast_group(PacketPimJoinPruneMulticastGroup(group, pruned_src_addresses=[source]))
            pckt = Packet(payload=PacketPimHeader(ph))

            self.get_interface().send(pckt.bytes())
            print("send prune echo")
        except:
            traceback.print_exc()
            return

    def send_join(self):
        print("send join")

        try:
            (source, group) = self.get_tree_id()
            ph = PacketPimJoinPrune(self.get_neighbor_RPF(), 210)
            ph.add_multicast_group(PacketPimJoinPruneMulticastGroup(group, joined_src_addresses=[source]))
            pckt = Packet(payload=PacketPimHeader(ph))

            self.get_interface().send(pckt.bytes())
        except:
            traceback.print_exc()
            return

    def send_assert(self):
        print("send assert")

        try:
            (source, group) = self.get_tree_id()
            assert_metric = self.my_assert_metric()
            ph = PacketPimAssert(multicast_group_address=group, source_address=source, metric_preference=assert_metric.metric_preference, metric=assert_metric.route_metric)
            pckt = Packet(payload=PacketPimHeader(ph))

            self.get_interface().send(pckt.bytes())
        except:
            traceback.print_exc()
            return

    def send_assert_cancel(self):
        print("send assert cancel")

        try:
            (source, group) = self.get_tree_id()
            ph = PacketPimAssert(multicast_group_address=group, source_address=source, metric_preference=float("Inf"), metric=float("Inf"))
            pckt = Packet(payload=PacketPimHeader(ph))

            self.get_interface().send(pckt.bytes())
        except:
            traceback.print_exc()
            return

    def send_state_refresh(self, state_refresh_msg_received: PacketPimStateRefresh):
        pass

    #############################################################

    @abstractmethod
    def is_forwarding(self):
        pass

    def assert_winner_nlt_expires(self):
        self._assert_state.winnerLivelinessTimerExpires(self)

    @abstractmethod
    def new_or_reset_neighbor(self, neighbor_ip):
        raise NotImplementedError()

    @abstractmethod
    def delete(self, change_type_interface=False):
        if change_type_interface:
            if self.could_assert():
                self._assert_state.couldAssertIsNowFalse(self)
            else:
                self._assert_state.couldAssertIsNowTrue(self)

        (s, g) = self.get_tree_id()
        # unsubscribe igmp information
        try:
            membership_interface = self.get_membership_interface()
            group_state = membership_interface.interface_state.get_group_state(g)
            group_state.remove_multicast_routing_entry(self)
        except:
            pass

        # Prune State
        self._prune_state = None

        # Assert State
        self._assert_state = None
        self.set_assert_winner_metric(AssertMetric.infinite_assert_metric()) # unsubscribe from current AssertWinner NeighborLivenessTimer
        self._assert_winner_metric = None
        self.clear_assert_timer()

        print('Tree Interface deleted')

    def is_olist_null(self):
        return self._kernel_entry.is_olist_null()

    def evaluate_ingroup(self):
        self._kernel_entry.evaluate_olist_change()

    #############################################################
    # Local Membership (IGMP)
    ############################################################
    def notify_membership(self, has_members: bool):
        with self.get_state_lock():
            with self._membership_lock:
                if has_members != self._local_membership_state.has_members():
                    self._local_membership_state = LocalMembership.Include if has_members else LocalMembership.NoInfo
                    self.change_tree()
                    self.evaluate_ingroup()

    def igmp_has_members(self):
        with self._membership_lock:
            return self._local_membership_state.has_members()

    def get_interface_name(self):
        return self._kernel_entry.get_interface_name(self._interface_id)

    def get_interface(self):
        return self._kernel_entry.get_interface(self._interface_id)

    def get_membership_interface(self):
        return self._kernel_entry.get_membership_interface(self._interface_id)

    def get_ip(self):
        ip = self.get_interface().get_ip()
        return ip

    def has_neighbors(self):
        try:
            return len(self.get_interface().neighbors) > 0
        except:
            return False

    def get_tree_id(self):
        return (self._kernel_entry.source_ip, self._kernel_entry.group_ip)

    def change_tree(self):
        self._kernel_entry.change()

    def get_state_lock(self):
        return self._kernel_entry.CHANGE_STATE_LOCK

    @abstractmethod
    def is_downstream(self):
        raise NotImplementedError()

    # obtain ip of RPF'(S)
    def get_neighbor_RPF(self):
        '''
        RPF'(S)
        '''
        if self.i_am_assert_loser():
            return self._assert_winner_metric.get_ip()
        else:
            return self._kernel_entry.rpf_node

    def is_S_directly_conn(self):
        return self._kernel_entry.rpf_node == self._kernel_entry.source_ip

    def set_receceived_prune_holdtime(self, holdtime):
        self._received_prune_holdtime = holdtime

    def get_received_prune_holdtime(self):
        return self._received_prune_holdtime

    ###################################################
    # ASSERT
    ###################################################
    def lost_assert(self):
        if not self.is_downstream():
            return False
        else:
            return not self._assert_winner_metric.i_am_assert_winner(self) and \
                   self._assert_winner_metric.is_better_than(AssertMetric.spt_assert_metric(self))

    def i_am_assert_loser(self):
        return self._assert_state == AssertState.Loser

    def could_assert(self):
        return self.is_downstream()

    def my_assert_metric(self):
        '''
        The assert metric of this interface for usage in assert state machine
        @rtype: AssertMetric
        '''
        if self.could_assert():
            return AssertMetric.spt_assert_metric(self)
        else:
            return AssertMetric.infinite_assert_metric()



import ipaddress


class AssertMetric(object):
    def __init__(self, metric_preference: int = 0x7FFFFFFF, route_metric: int = 0xFFFFFFFF, ip_address: str = "0.0.0.0", state_refresh_interval:int = None):
        if type(ip_address) is str:
            ip_address = ipaddress.ip_address(ip_address)

        self._metric_preference = metric_preference
        self._route_metric = route_metric
        self._ip_address = ip_address
        self._state_refresh_interval = state_refresh_interval

    def is_better_than(self, other):
        if self.metric_preference != other.metric_preference:
            return self.metric_preference < other.metric_preference
        elif self.route_metric != other.route_metric:
            return self.route_metric < other.route_metric
        else:
            return self.ip_address > other.ip_address

    def is_worse(self, other):
        return not self.is_better_than(other)

    def equal_metric(self, other):
        return self.metric_preference == other.metric_preference and self.metric_preference == other.metric_preference \
               and self.ip_address == other.ip_address

    @staticmethod
    def infinite_assert_metric():
        '''
        @type metric: AssertMetric
        '''
        return AssertMetric()

    @staticmethod
    def spt_assert_metric(tree_if):
        '''
        @type metric: AssertMetric
        @type tree_if: TreeInterface
        '''
        (source_ip, _) = tree_if.get_tree_id()
        from pimdm import UnicastRouting
        (metric_preference, metric_cost, _) = UnicastRouting.get_metric(source_ip)
        return AssertMetric(metric_preference, metric_cost, tree_if.get_ip())


    def i_am_assert_winner(self, tree_if):
        return self.get_ip() == tree_if.get_ip()

    @property
    def metric_preference(self):
        return self._metric_preference

    @metric_preference.setter
    def metric_preference(self, value):
        self._metric_preference = value

    @property
    def route_metric(self):
        return self._route_metric

    @route_metric.setter
    def route_metric(self, value):
        self._route_metric = value


    @property
    def ip_address(self):
        return self._ip_address

    @ip_address.setter
    def ip_address(self, value):
        if type(value) is str:
            value = ipaddress.ip_address(value)

        self._ip_address = value

    @property
    def state_refresh_interval(self):
        return self._state_refresh_interval

    @state_refresh_interval.setter
    def state_refresh_interval(self, value):
        self._state_refresh_interval = value

    def get_ip(self):
        return str(self._ip_address)



from abc import ABCMeta, abstractmethod

from pimdm.tree import pim_globals
from .metric import AssertMetric
from pimdm.utils import TYPE_CHECKING
if TYPE_CHECKING:
    from .tree_if_downstream import TreeInterfaceDownstream


class AssertStateABC(metaclass=ABCMeta):
    @staticmethod
    @abstractmethod
    def receivedDataFromDownstreamIf(interface: "TreeInterfaceDownstream"):
        """
        An (S,G) Data packet received on downstream interface

        @type interface: TreeInterface
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def receivedInferiorMetricFromWinner(interface: "TreeInterfaceDownstream"):
        """
        Receive Inferior (Assert OR State Refresh) from Assert Winner

        @type interface: TreeInterface
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def receivedInferiorMetricFromNonWinner_couldAssertIsTrue(interface: "TreeInterfaceDownstream"):
        """
        Receive Inferior (Assert OR  State Refresh) from non-Assert Winner
        AND CouldAssert==TRUE

        @type interface: TreeInterface
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def receivedPreferedMetric(interface: "TreeInterfaceDownstream", better_metric, is_metric_equal):
        """
        Receive Preferred Assert OR State Refresh

        @type interface: TreeInterface
        @type better_metric: AssertMetric
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def sendStateRefresh(interface: "TreeInterfaceDownstream", time):
        """
        Send State Refresh


        @type interface: TreeInterface
        @type time: int
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def assertTimerExpires(interface: "TreeInterfaceDownstream"):
        """
        AT(S,G) Expires

        @type interface: TreeInterface
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def couldAssertIsNowFalse(interface: "TreeInterfaceDownstream"):
        """
        CouldAssert -> FALSE

        @type interface: TreeInterface
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def couldAssertIsNowTrue(interface: "TreeInterfaceDownstream"):
        """
        CouldAssert -> TRUE

        @type interface: TreeInterface
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def winnerLivelinessTimerExpires(interface: "TreeInterfaceDownstream"):
        """
        Winner’s NLT(N,I) Expires

        @type interface: TreeInterface
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def receivedPruneOrJoinOrGraft(interface: "TreeInterfaceDownstream"):
        """
        Receive Prune(S,G), Join(S,G) or Graft(S,G)

        @type interface: TreeInterface
        """
        raise NotImplementedError()

    def _sendAssert_setAT(interface: "TreeInterfaceDownstream"):
        interface.set_assert_timer(pim_globals.ASSERT_TIME)
        interface.send_assert()

    @staticmethod
    @abstractmethod
    def is_preferred_assert(interface: "TreeInterfaceDownstream", received_metric):
        raise NotImplementedError()

    # Override
    def __str__(self) -> str:
        return "AssertSM:" + self.__class__.__name__


class NoInfoState(AssertStateABC):
    '''
    NoInfoState (NI)
    This router has no (S,G) Assert state on interface I.
    '''

    @staticmethod
    def is_preferred_assert(interface: "TreeInterfaceDownstream", received_metric):
        return received_metric.is_better_than(interface._assert_winner_metric)

    @staticmethod
    def receivedDataFromDownstreamIf(interface: "TreeInterfaceDownstream"):
        """
        @type interface: TreeInterface
        """
        interface.assert_logger.debug('receivedDataFromDownstreamIf, NI -> W')

        interface.set_assert_winner_metric(interface.my_assert_metric())
        interface.set_assert_state(AssertState.Winner)

        NoInfoState._sendAssert_setAT(interface)

    @staticmethod
    def receivedInferiorMetricFromWinner(interface: "TreeInterfaceDownstream"):
        assert False, "this should never ocurr"

    @staticmethod
    def receivedInferiorMetricFromNonWinner_couldAssertIsTrue(interface: "TreeInterfaceDownstream"):
        interface.assert_logger.debug('receivedInferiorMetricFromNonWinner_couldAssertIsTrue, NI -> W')

        interface.set_assert_winner_metric(interface.my_assert_metric())
        interface.set_assert_state(AssertState.Winner)

        NoInfoState._sendAssert_setAT(interface)

    @staticmethod
    def receivedPreferedMetric(interface: "TreeInterfaceDownstream", better_metric, is_metric_equal):
        '''
        @type interface: TreeInterface
        '''
        if is_metric_equal:
            return
        interface.assert_logger.debug('receivedPreferedMetric, NI -> L')
        state_refresh_interval = better_metric.state_refresh_interval
        if state_refresh_interval is None:
            # event caused by Assert Msg
            assert_timer_value = pim_globals.ASSERT_TIME
        else:
            # event caused by StateRefreshMsg
            assert_timer_value = state_refresh_interval*3

        interface.set_assert_timer(assert_timer_value)
        interface.set_assert_winner_metric(better_metric)
        interface.set_assert_state(AssertState.Loser)

        # MUST also multicast a Prune(S,G) to the Assert winner
        if interface.could_assert():
            interface.send_prune(holdtime=assert_timer_value)

    @staticmethod
    def sendStateRefresh(interface: "TreeInterfaceDownstream", time):
        pass

    @staticmethod
    def assertTimerExpires(interface: "TreeInterfaceDownstream"):
        assert False, "this should never ocurr"

    @staticmethod
    def couldAssertIsNowFalse(interface: "TreeInterfaceDownstream"):
        interface.assert_logger.debug('couldAssertIsNowFalse, NI -> NI')

    @staticmethod
    def couldAssertIsNowTrue(interface: "TreeInterfaceDownstream"):
        interface.assert_logger.debug('couldAssertIsNowTrue, NI -> NI')

    @staticmethod
    def winnerLivelinessTimerExpires(interface: "TreeInterfaceDownstream"):
        assert False, "this should never ocurr"

    @staticmethod
    def receivedPruneOrJoinOrGraft(interface: "TreeInterfaceDownstream"):
        interface.assert_logger.debug('receivedPruneOrJoinOrGraft, NI -> NI')

    def __str__(self) -> str:
        return "NoInfo"


class WinnerState(AssertStateABC):
    '''
    I am Assert Winner (W)
    This router has won an (S,G) Assert on interface I. It is now
    responsible for forwarding traffic from S destined for G via
    interface I.
    '''

    @staticmethod
    def is_preferred_assert(interface: "TreeInterfaceDownstream", received_metric):
        return received_metric.is_better_than(interface.my_assert_metric())

    @staticmethod
    def receivedDataFromDownstreamIf(interface: "TreeInterfaceDownstream"):
        """
        @type interface: TreeInterface
        """
        interface.assert_logger.debug('receivedDataFromDownstreamIf, W -> W')
        WinnerState._sendAssert_setAT(interface)

    @staticmethod
    def receivedInferiorMetricFromWinner(interface: "TreeInterfaceDownstream"):
        assert False, "this should never ocurr"

    @staticmethod
    def receivedInferiorMetricFromNonWinner_couldAssertIsTrue(interface: "TreeInterfaceDownstream"):
        interface.assert_logger.debug('receivedInferiorMetricFromNonWinner_couldAssertIsTrue, W -> W')
        WinnerState._sendAssert_setAT(interface)

    @staticmethod
    def receivedPreferedMetric(interface: "TreeInterfaceDownstream", better_metric, is_metric_equal):
        '''
        @type better_metric: AssertMetric
        '''
        if is_metric_equal:
            return
        interface.assert_logger.debug('receivedPreferedMetric, W -> L')
        state_refresh_interval = better_metric.state_refresh_interval
        if state_refresh_interval is None:
            # event caused by AssertMsg
            assert_timer_value = pim_globals.ASSERT_TIME
        else:
            # event caused by State Refresh Msg
            assert_timer_value = state_refresh_interval*3

        interface.set_assert_timer(assert_timer_value)
        interface.set_assert_winner_metric(better_metric)
        interface.set_assert_state(AssertState.Loser)

        interface.send_prune(holdtime=assert_timer_value)

    @staticmethod
    def sendStateRefresh(interface: "TreeInterfaceDownstream", state_refresh_interval):
        interface.set_assert_timer(state_refresh_interval*3)

    @staticmethod
    def assertTimerExpires(interface: "TreeInterfaceDownstream"):
        interface.assert_logger.debug('assertTimerExpires, W -> NI')
        interface.set_assert_winner_metric(AssertMetric.infinite_assert_metric())
        interface.set_assert_state(AssertState.NoInfo)

    @staticmethod
    def couldAssertIsNowFalse(interface: "TreeInterfaceDownstream"):
        interface.assert_logger.debug('couldAssertIsNowFalse, W -> NI')
        interface.send_assert_cancel()

        interface.clear_assert_timer()

        interface.set_assert_winner_metric(AssertMetric.infinite_assert_metric())
        interface.set_assert_state(AssertState.NoInfo)

    @staticmethod
    def couldAssertIsNowTrue(interface: "TreeInterfaceDownstream"):
        assert False, "this should never ocurr"

    @staticmethod
    def winnerLivelinessTimerExpires(interface: "TreeInterfaceDownstream"):
        assert False, "this should never ocurr"

    @staticmethod
    def receivedPruneOrJoinOrGraft(interface: "TreeInterfaceDownstream"):
        pass

    def __str__(self) -> str:
        return "Winner"


class LoserState(AssertStateABC):
    '''
    I am Assert Loser (L)
    This router has lost an (S,G) Assert on interface I. It must not
    forward packets from S destined for G onto interface I.
    '''

    @staticmethod
    def is_preferred_assert(interface: "TreeInterfaceDownstream", received_metric):
        return received_metric.is_better_than(interface._assert_winner_metric) or \
               received_metric.equal_metric(interface._assert_winner_metric)

    @staticmethod
    def receivedDataFromDownstreamIf(interface: "TreeInterfaceDownstream"):
        """
        @type interface: TreeInterface
        """
        interface.assert_logger.debug('receivedDataFromDownstreamIf, L -> L')

    @staticmethod
    def receivedInferiorMetricFromWinner(interface: "TreeInterfaceDownstream"):
        interface.assert_logger.debug('receivedInferiorMetricFromWinner, L -> NI')
        LoserState._to_NoInfo(interface)

    @staticmethod
    def receivedInferiorMetricFromNonWinner_couldAssertIsTrue(interface: "TreeInterfaceDownstream"):
        interface.assert_logger.debug('receivedInferiorMetricFromNonWinner_couldAssertIsTrue, L -> L')

    @staticmethod
    def receivedPreferedMetric(interface: "TreeInterfaceDownstream", better_metric, is_metric_equal):
        '''
        @type better_metric: AssertMetric
        '''
        interface.assert_logger.debug('receivedPreferedMetric, L -> L')
        state_refresh_interval = better_metric.state_refresh_interval
        if state_refresh_interval is None:
            assert_timer_value = pim_globals.ASSERT_TIME
        else:
            assert_timer_value = state_refresh_interval*3

        interface.set_assert_timer(assert_timer_value)
        interface.set_assert_winner_metric(better_metric)
        interface.set_assert_state(AssertState.Loser)

        if not is_metric_equal and interface.could_assert():
            interface.send_prune(holdtime=assert_timer_value)

    @staticmethod
    def sendStateRefresh(interface: "TreeInterfaceDownstream", time):
        assert False, "this should never ocurr"

    @staticmethod
    def assertTimerExpires(interface: "TreeInterfaceDownstream"):
        interface.assert_logger.debug('assertTimerExpires, L -> NI')
        LoserState._to_NoInfo(interface)

    @staticmethod
    def couldAssertIsNowFalse(interface: "TreeInterfaceDownstream"):
        interface.assert_logger.debug('couldAssertIsNowFalse, L -> NI')
        LoserState._to_NoInfo(interface)

    @staticmethod
    def couldAssertIsNowTrue(interface: "TreeInterfaceDownstream"):
        interface.assert_logger.debug('couldAssertIsNowTrue, L -> NI')
        LoserState._to_NoInfo(interface)

    @staticmethod
    def winnerLivelinessTimerExpires(interface: "TreeInterfaceDownstream"):
        interface.assert_logger.debug('winnerLivelinessTimerExpires, L -> NI')
        LoserState._to_NoInfo(interface)

    @staticmethod
    def receivedPruneOrJoinOrGraft(interface: "TreeInterfaceDownstream"):
        interface.assert_logger.debug('receivedPruneOrJoinOrGraft, L -> L')
        interface.send_assert()

    @staticmethod
    def _to_NoInfo(interface: "TreeInterfaceDownstream"):
        interface.clear_assert_timer()
        interface.set_assert_winner_metric(AssertMetric.infinite_assert_metric())
        interface.set_assert_state(AssertState.NoInfo)

    def __str__(self) -> str:
        return "Loser"


class AssertState():
    NoInfo = NoInfoState()
    Winner = WinnerState()
    Loser = LoserState()



import logging

class RootFilter(logging.Filter):
    """
    This is a filter which injects contextual information into the log.

    Rather than use actual contextual information, we just use random
    data in this demo.
    """
    def __init__(self, router_name):
        super().__init__()
        self.router_name = router_name

    def filter(self, record):
        record.routername = self.router_name
        if not hasattr(record, 'tree'):
            record.tree = ''
        if not hasattr(record, 'vif'):
            record.vif = ''
        if not hasattr(record, 'interfacename'):
            record.interfacename = ''
        if not hasattr(record, 'neighbor_ip'):
            record.neighbor_ip = ''
        return True



import socket
import random
import struct
import logging
import ipaddress
import netifaces
from pimdm import Main
from socket import if_nametoindex
from pimdm.Interface import Interface
from .InterfacePIM import InterfacePim
from pimdm.rwlock.RWLock import RWLockWrite
from pimdm.packet.ReceivedPacket import ReceivedPacket_v6


class InterfacePim6(InterfacePim):
    MCAST_GRP = "ff02::d"

    def __init__(self, interface_name: str, vif_index:int, state_refresh_capable:bool=False):
        # generation id
        self.generation_id = random.getrandbits(32)

        # When PIM is enabled on an interface or when a router first starts, the Hello Timer (HT)
        # MUST be set to random value between 0 and Triggered_Hello_Delay
        self.hello_timer = None

        # state refresh capable
        self._state_refresh_capable = state_refresh_capable
        self._neighbors_state_refresh_capable = False

        # todo: lan delay enabled
        self._lan_delay_enabled = False

        # todo: propagation delay
        self._propagation_delay = self.PROPAGATION_DELAY

        # todo: override interval
        self._override_interval = self.OVERRIDE_INTERNAL

        # pim neighbors
        self._had_neighbors = False
        self.neighbors = {}
        self.neighbors_lock = RWLockWrite()
        self.interface_logger = logging.LoggerAdapter(InterfacePim.LOGGER, {'vif': vif_index, 'interfacename': interface_name})

        # SOCKET
        s = socket.socket(socket.AF_INET6, socket.SOCK_RAW, socket.IPPROTO_PIM)

        ip_interface = ""
        for if_addr in netifaces.ifaddresses(interface_name)[netifaces.AF_INET6]:
            ip_interface = if_addr["addr"]
            if ipaddress.IPv6Address(if_addr['addr'].split("%")[0]).is_link_local:
                ip_interface = if_addr['addr'].split("%")[0]
                # bind to interface
                s.bind(socket.getaddrinfo(if_addr['addr'], None, 0, socket.SOCK_RAW, 0, socket.AI_PASSIVE)[0][4])
                break

        self.ip_interface = ip_interface

        # allow other sockets to bind this port too
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        # explicitly join the multicast group on the interface specified
        if_index = if_nametoindex(interface_name)
        s.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_JOIN_GROUP,
                     socket.inet_pton(socket.AF_INET6, InterfacePim6.MCAST_GRP) + struct.pack('@I', if_index))
        s.setsockopt(socket.SOL_SOCKET, 25, str(interface_name + '\0').encode('utf-8'))

        # set socket output interface
        s.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_MULTICAST_IF, struct.pack('@I', if_index))

        # set socket TTL to 1
        s.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_MULTICAST_HOPS, 1)
        s.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_UNICAST_HOPS, 1)

        # don't receive outgoing packets
        s.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_MULTICAST_LOOP, 0)
        Interface.__init__(self, interface_name, s, s, vif_index)
        Interface._enable(self)
        self.force_send_hello()

    @staticmethod
    def get_kernel():
        return Main.kernel_v6

    def send(self, data: bytes, group_ip: str=MCAST_GRP):
        super().send(data=data, group_ip=group_ip)

    def _receive(self, raw_bytes, ancdata, src_addr):
        if raw_bytes:
            packet = ReceivedPacket_v6(raw_bytes, ancdata, src_addr, 103, self)
            self.PKT_FUNCTIONS[packet.payload.get_pim_type()](self, packet)



from abc import ABCMeta, abstractmethod
from pimdm.utils import TYPE_CHECKING
if TYPE_CHECKING:
    from .tree_if_upstream import TreeInterfaceUpstream

class UpstreamStateABC(metaclass=ABCMeta):
    @staticmethod
    @abstractmethod
    def dataArrivesRPFinterface_OListNull_PLTstoped(interface: "TreeInterfaceUpstream"):
        """
        Data arrives on RPF_Interface(S) AND
        olist(S, G) == NULL AND
        PLT(S, G) not running

        @type interface: Upstream
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def stateRefreshArrivesRPFnbr_pruneIs1(interface: "TreeInterfaceUpstream"):
        """
        State Refresh(S,G) received from RPF‘(S) AND
        Prune Indicator == 1

        @type interface: Upstream
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def stateRefreshArrivesRPFnbr_pruneIs0_PLTstoped(interface: "TreeInterfaceUpstream"):
        """
        State Refresh(S,G) received from RPF‘(S) AND
        Prune Indicator == 0 AND
        PLT(S, G) not running

        @type interface: Upstream
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def seeJoinToRPFnbr(interface: "TreeInterfaceUpstream"):
        """
        See Join(S,G) to RPF’(S)

        @type interface: Upstream
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def seePrune(interface: "TreeInterfaceUpstream"):
        """
        See Prune(S,G)

        @type interface: Upstream
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def OTexpires(interface: "TreeInterfaceUpstream"):
        """
        OT(S,G) Expires

        @type interface: Upstream
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def olistIsNowNull(interface: "TreeInterfaceUpstream"):
        """
        olist(S,G)->NULL

        @type interface: Upstream
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def olistIsNowNotNull(interface: "TreeInterfaceUpstream"):
        """
        olist(S,G)->non-NULL

        @type interface: Upstream
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def RPFnbrChanges_olistIsNotNull(interface: "TreeInterfaceUpstream"):
        """
        RPF’(S) Changes AND
        olist(S,G) != NULL AND
        S not directly connected

        @type interface: Upstream
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def RPFnbrChanges_olistIsNull(interface: "TreeInterfaceUpstream"):
        """
        RPF’(S) Changes AND
        olist(S,G) == NULL

        @type interface: Upstream
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def sourceIsNowDirectConnect(interface: "TreeInterfaceUpstream"):
        """
        S becomes directly connected

        @type interface: Upstream
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def GRTexpires(interface: "TreeInterfaceUpstream"):
        """
        GRT(S,G) Expires

        @type interface: Upstream
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def recvGraftAckFromRPFnbr(interface: "TreeInterfaceUpstream"):
        """
        Receive GraftAck(S,G) from RPF’(S)

        @type interface: Upstream
        """
        raise NotImplementedError()


class Forward(UpstreamStateABC):
    """
    Forwarding (F)
    This is the starting state of the Upsteam(S,G) state machine.
    The state machine is in this state if it just started or if
    oiflist(S,G) != NULL.
    """

    @staticmethod
    def dataArrivesRPFinterface_OListNull_PLTstoped(interface: "TreeInterfaceUpstream"):
        """
        Data arrives on RPF_Interface(S) AND
        olist(S, G) == NULL AND
        PLT(S, G) not running

        @type interface: TreeInterfaceUpstream
        """
        if not interface.is_S_directly_conn():
            interface.join_prune_logger.debug("dataArrivesRPFinterface_OListNull_PLTstoped, F -> P")

            interface.set_state(UpstreamState.Pruned)
            interface.send_prune()
            interface.set_prune_limit_timer()

    @staticmethod
    def stateRefreshArrivesRPFnbr_pruneIs1(interface: "TreeInterfaceUpstream"):
        """
        State Refresh(S,G) received from RPF‘(S) AND
        Prune Indicator == 1

        @type interface: TreeInterfaceUpstream
        """
        # if OT is not running the router must set OT to t_override seconds
        interface.join_prune_logger.debug('stateRefreshArrivesRPFnbr_pruneIs1, F -> F')
        if not interface.is_override_timer_running():
            interface.set_override_timer()

    @staticmethod
    def stateRefreshArrivesRPFnbr_pruneIs0_PLTstoped(interface: "TreeInterfaceUpstream"):
        """
        State Refresh(S,G) received from RPF‘(S) AND
        Prune Indicator == 0 AND
        PLT(S, G) not running

        @type interface: TreeInterfaceUpstream
        """
        interface.join_prune_logger.debug('stateRefreshArrivesRPFnbr_pruneIs0_PLTstoped, F -> F')

    @staticmethod
    def seeJoinToRPFnbr(interface: "TreeInterfaceUpstream"):
        """
        See Join(S,G) to RPF’(S)

        @type interface: TreeInterfaceUpstream
        """
        interface.join_prune_logger.debug('seeJoinToRPFnbr, F -> F')
        interface.clear_override_timer()

    @staticmethod
    def seePrune(interface: "TreeInterfaceUpstream"):
        """
        See Prune(S,G)

        @type interface: TreeInterfaceUpstream
        """
        interface.join_prune_logger.debug('seePrune, F -> F')
        if not interface.is_S_directly_conn() and not interface.is_override_timer_running():
            interface.set_override_timer()

    @staticmethod
    def OTexpires(interface: "TreeInterfaceUpstream"):
        """
        OT(S,G) Expires

        @type interface: TreeInterfaceUpstream
        """
        interface.join_prune_logger.debug('OTexpires, F -> F')
        if not interface.is_S_directly_conn():
            interface.send_join()

    @staticmethod
    def olistIsNowNull(interface: "TreeInterfaceUpstream"):
        """
        olist(S,G)->NULL

        @type interface: TreeInterfaceUpstream
        """
        if not interface.is_S_directly_conn():
            interface.join_prune_logger.debug("olistIsNowNull, F -> P")

            interface.set_state(UpstreamState.Pruned)

            interface.send_prune()
            interface.set_prune_limit_timer()

    @staticmethod
    def olistIsNowNotNull(interface: "TreeInterfaceUpstream"):
        """
        olist(S,G)->non-NULL

        @type interface: TreeInterfaceUpstream
        """
        #assert False, "olistIsNowNotNull (in state F)"
        return

    @staticmethod
    def RPFnbrChanges_olistIsNotNull(interface: "TreeInterfaceUpstream"):
        """
        RPF’(S) Changes AND
        olist(S,G) != NULL AND
        S not directly connected

        @type interface: TreeInterfaceUpstream
        """
        if not interface.is_S_directly_conn():
            interface.join_prune_logger.debug('RPFnbrChanges_olistIsNotNull, F -> AP')

            interface.set_state(UpstreamState.AckPending)

            interface.send_graft()
            interface.set_graft_retry_timer()

    @staticmethod
    def RPFnbrChanges_olistIsNull(interface: "TreeInterfaceUpstream"):
        """
        RPF’(S) Changes AND
        olist(S,G) == NULL

        @type interface: TreeInterfaceUpstream
        """
        interface.join_prune_logger.debug('RPFnbrChanges_olistIsNull, F -> P')
        interface.set_state(UpstreamState.Pruned)

    @staticmethod
    def sourceIsNowDirectConnect(interface: "TreeInterfaceUpstream"):
        """
        S becomes directly connected

        @type interface: TreeInterfaceUpstream
        """
        interface.join_prune_logger.debug("sourceIsNowDirectConnect, F -> F")

    @staticmethod
    def GRTexpires(interface: "TreeInterfaceUpstream"):
        """
        GRT(S,G) Expires

        @type interface: TreeInterfaceUpstream
        """
        #assert False, "GRTexpires (in state F)"
        return

    @staticmethod
    def recvGraftAckFromRPFnbr(interface: "TreeInterfaceUpstream"):
        """
        Receive GraftAck(S,G) from RPF’(S)

        @type interface: TreeInterfaceUpstream
        """
        interface.join_prune_logger.debug("recvGraftAckFromRPFnbr, F -> F")

    def __str__(self):
        return "Forwarding"


class Pruned(UpstreamStateABC):
    '''
    Pruned (P)
    The set, olist(S,G), is empty.
    The router will not forward data from S addressed to group G.
    '''

    @staticmethod
    def dataArrivesRPFinterface_OListNull_PLTstoped(interface: "TreeInterfaceUpstream"):
        """
        Data arrives on RPF_Interface(S) AND
        olist(S, G) == NULL AND
        PLT(S, G) not running

        @type interface: TreeInterfaceUpstream
        """
        if not interface.is_S_directly_conn():
            interface.join_prune_logger.debug("dataArrivesRPFinterface_OListNull_PLTstoped, P -> P")
            interface.set_prune_limit_timer()
            interface.send_prune()

    @staticmethod
    def stateRefreshArrivesRPFnbr_pruneIs1(interface: "TreeInterfaceUpstream"):
        """
        State Refresh(S,G) received from RPF‘(S) AND
        Prune Indicator == 1

        @type interface: TreeInterfaceUpstream
        """
        interface.join_prune_logger.debug('stateRefreshArrivesRPFnbr_pruneIs1, P -> P')
        interface.set_prune_limit_timer()

    @staticmethod
    def stateRefreshArrivesRPFnbr_pruneIs0_PLTstoped(interface: "TreeInterfaceUpstream"):
        """
        State Refresh(S,G) received from RPF‘(S) AND
        Prune Indicator == 0 AND
        PLT(S, G) not running

        @type interface: TreeInterfaceUpstream
        """
        interface.join_prune_logger.debug('stateRefreshArrivesRPFnbr_pruneIs0_PLTstoped, P -> P')
        interface.send_prune()
        interface.set_prune_limit_timer()

    @staticmethod
    def seeJoinToRPFnbr(interface: "TreeInterfaceUpstream"):
        """
        See Join(S,G) to RPF’(S)

        @type interface: TreeInterfaceUpstream
        """
        # Do nothing
        interface.join_prune_logger.debug('seeJoinToRPFnbr, P -> P')

    @staticmethod
    def seePrune(interface: "TreeInterfaceUpstream"):
        """
        See Prune(S,G)

        @type interface: TreeInterfaceUpstream
        """
        interface.join_prune_logger.debug('seePrune, P -> P')
        if interface.get_received_prune_holdtime() > interface.remaining_prune_limit_timer():
            interface.set_prune_limit_timer(time=interface.get_received_prune_holdtime())

    @staticmethod
    def OTexpires(interface: "TreeInterfaceUpstream"):
        """
        OT(S,G) Expires

        @type interface: TreeInterfaceUpstream
        """
        #assert False, "OTexpires in state Pruned"
        return

    @staticmethod
    def olistIsNowNull(interface: "TreeInterfaceUpstream"):
        """
        olist(S,G)->NULL

        @type interface: TreeInterfaceUpstream
        """
        #assert False, "olistIsNowNull in state Pruned"
        return

    @staticmethod
    def olistIsNowNotNull(interface: "TreeInterfaceUpstream"):
        """
        olist(S,G)->non-NULL

        @type interface: TreeInterfaceUpstream
        """
        if not interface.is_S_directly_conn():
            interface.join_prune_logger.debug('olistIsNowNotNull, P -> AP')

            interface.clear_prune_limit_timer()

            interface.set_state(UpstreamState.AckPending)

            interface.send_graft()
            interface.set_graft_retry_timer()

    @staticmethod
    def RPFnbrChanges_olistIsNotNull(interface: "TreeInterfaceUpstream"):
        """
        RPF’(S) Changes AND
        olist(S,G) != NULL AND
        S not directly connected

        @type interface: TreeInterfaceUpstream
        """
        if not interface.is_S_directly_conn():
            interface.join_prune_logger.debug('RPFnbrChanges_olistIsNotNull, P -> AP')

            interface.clear_prune_limit_timer()

            interface.set_state(UpstreamState.AckPending)

            interface.send_graft()
            interface.set_graft_retry_timer()

    @staticmethod
    def RPFnbrChanges_olistIsNull(interface: "TreeInterfaceUpstream"):
        """
        RPF’(S) Changes AND
        olist(S,G) == NULL

        @type interface: TreeInterfaceUpstream
        """
        if not interface.is_S_directly_conn():
            interface.join_prune_logger.debug('RPFnbrChanges_olistIsNull, P -> P')
            interface.clear_prune_limit_timer()

    @staticmethod
    def sourceIsNowDirectConnect(interface: "TreeInterfaceUpstream"):
        """
        S becomes directly connected

        @type interface: TreeInterfaceUpstream
        """
        interface.join_prune_logger.debug('sourceIsNowDirectConnect, P -> P')

    @staticmethod
    def GRTexpires(interface: "TreeInterfaceUpstream"):
        """
        GRT(S,G) Expires

        @type interface: TreeInterfaceUpstream
        """
        #assert False, "GRTexpires in state Pruned"
        return

    @staticmethod
    def recvGraftAckFromRPFnbr(interface: "TreeInterfaceUpstream"):
        """
        Receive GraftAck(S,G) from RPF’(S)

        @type interface: TreeInterfaceUpstream
        """
        interface.join_prune_logger.debug('recvGraftAckFromRPFnbr, P -> P')

    def __str__(self):
        return "Pruned"


class AckPending(UpstreamStateABC):
    """
    AckPending (AP)
    The router was in the Pruned(P) state, but a transition has
    occurred in the Downstream(S,G) state machine for one of this
    (S,G) entry’s outgoing interfaces, indicating that traffic from S
    addressed to G should again be forwarded. A Graft message has
    been sent to RPF’(S), but a Graft Ack message has not yet been
    received.
    """

    @staticmethod
    def dataArrivesRPFinterface_OListNull_PLTstoped(interface: "TreeInterfaceUpstream"):
        """
        Data arrives on RPF_Interface(S) AND
        olist(S, G) == NULL AND
        PLT(S, G) not running

        @type interface: TreeInterfaceUpstream
        """
        #assert False, "dataArrivesRPFinterface_OListNull_PLTstoped in state AP"
        return

    @staticmethod
    def stateRefreshArrivesRPFnbr_pruneIs1(interface: "TreeInterfaceUpstream"):
        """
        State Refresh(S,G) received from RPF‘(S) AND
        Prune Indicator == 1

        @type interface: TreeInterfaceUpstream
        """
        interface.join_prune_logger.debug('stateRefreshArrivesRPFnbr_pruneIs1, AP -> AP')
        if not interface.is_override_timer_running():
            interface.set_override_timer()

    @staticmethod
    def stateRefreshArrivesRPFnbr_pruneIs0_PLTstoped(interface: "TreeInterfaceUpstream"):
        """
        State Refresh(S,G) received from RPF‘(S) AND
        Prune Indicator == 0 AND
        PLT(S, G) not running

        @type interface: TreeInterfaceUpstream
        """
        interface.join_prune_logger.debug('stateRefreshArrivesRPFnbr_pruneIs0_PLTstoped, AP -> F')

        interface.clear_graft_retry_timer()
        interface.set_state(UpstreamState.Forward)

    @staticmethod
    def seeJoinToRPFnbr(interface: "TreeInterfaceUpstream"):
        """
        See Join(S,G) to RPF’(S)

        @type interface: TreeInterfaceUpstream
        """
        interface.join_prune_logger.debug('seeJoinToRPFnbr, AP -> AP')
        interface.clear_override_timer()

    @staticmethod
    def seePrune(interface: "TreeInterfaceUpstream"):
        """
        See Prune(S,G)

        @type interface: TreeInterfaceUpstream
        """
        interface.join_prune_logger.debug('seePrune, AP -> AP')
        if not interface.is_override_timer_running():
            interface.set_override_timer()

    @staticmethod
    def OTexpires(interface: "TreeInterfaceUpstream"):
        """
        OT(S,G) Expires

        @type interface: TreeInterfaceUpstream
        """
        interface.join_prune_logger.debug('OTexpires, AP -> AP')
        interface.send_join()

    @staticmethod
    def olistIsNowNull(interface: "TreeInterfaceUpstream"):
        """
        olist(S,G)->NULL

        @type interface: TreeInterfaceUpstream
        """
        interface.join_prune_logger.debug('olistIsNowNull, AP -> P')
        interface.set_state(UpstreamState.Pruned)

        interface.send_prune()

        interface.clear_graft_retry_timer()
        interface.set_prune_limit_timer()

    @staticmethod
    def olistIsNowNotNull(interface: "TreeInterfaceUpstream"):
        """
        olist(S,G)->non-NULL

        @type interface: TreeInterfaceUpstream
        """
        #assert False, "olistIsNowNotNull in state AP"
        return

    @staticmethod
    def RPFnbrChanges_olistIsNotNull(interface: "TreeInterfaceUpstream"):
        """
        RPF’(S) Changes AND
        olist(S,G) != NULL AND
        S not directly connected

        @type interface: TreeInterfaceUpstream
        """
        if not interface.is_S_directly_conn():
            interface.join_prune_logger.debug('RPFnbrChanges_olistIsNotNull, AP -> AP')

            interface.send_graft()
            interface.set_graft_retry_timer()

    @staticmethod
    def RPFnbrChanges_olistIsNull(interface: "TreeInterfaceUpstream"):
        """
        RPF’(S) Changes AND
        olist(S,G) == NULL

        @type interface: TreeInterfaceUpstream
        """
        if not interface.is_S_directly_conn():
            interface.join_prune_logger.debug('RPFnbrChanges_olistIsNull, AP -> P')

            interface.clear_graft_retry_timer()
            interface.set_state(UpstreamState.Pruned)

    @staticmethod
    def sourceIsNowDirectConnect(interface: "TreeInterfaceUpstream"):
        """
        S becomes directly connected

        @type interface: TreeInterfaceUpstream
        """
        interface.join_prune_logger.debug('sourceIsNowDirectConnect, AP -> F')

        interface.set_state(UpstreamState.Forward)
        interface.clear_graft_retry_timer()

    @staticmethod
    def GRTexpires(interface: "TreeInterfaceUpstream"):
        """
        GRT(S,G) Expires

        @type interface: TreeInterfaceUpstream
        """
        interface.join_prune_logger.debug('GRTexpires, AP -> AP')

        interface.set_graft_retry_timer()
        interface.send_graft()

    @staticmethod
    def recvGraftAckFromRPFnbr(interface: "TreeInterfaceUpstream"):
        """
        Receive GraftAck(S,G) from RPF’(S)

        @type interface: TreeInterfaceUpstream
        """
        interface.join_prune_logger.debug('recvGraftAckFromRPFnbr, AP -> F')

        interface.clear_graft_retry_timer()
        interface.set_state(UpstreamState.Forward)

    def __str__(self):
        return "AckPending"

class UpstreamState():
    Forward = Forward()
    Pruned = Pruned()
    AckPending = AckPending()



from pimdm import Main
from abc import abstractmethod, ABCMeta


class KernelEntryInterface(metaclass=ABCMeta):
    @staticmethod
    @abstractmethod
    def get_outbound_interfaces_indexes(kernel_tree):
        """
        Get OIL of this tree
        """
        pass

    @staticmethod
    @abstractmethod
    def get_interface_name(interface_id):
        """
        Get name of interface from vif id
        """
        pass

    @staticmethod
    @abstractmethod
    def get_interface(kernel_tree, interface_id):
        """
        Get PIM interface from interface id
        """
        pass

    @staticmethod
    @abstractmethod
    def get_membership_interface(kernel_tree, interface_id):
        """
        Get IGMP/MLD interface from interface id
        """
        pass

    @staticmethod
    @abstractmethod
    def get_kernel():
        """
        Get kernel
        """
        pass


class KernelEntry4Interface(KernelEntryInterface):
    @staticmethod
    def get_outbound_interfaces_indexes(kernel_tree):
        """
        Get OIL of this tree
        """
        outbound_indexes = [0] * Main.kernel.MAXVIFS
        for (index, state) in kernel_tree.interface_state.items():
            outbound_indexes[index] = state.is_forwarding()
        return outbound_indexes

    @staticmethod
    def get_interface_name(interface_id):
        """
        Get name of interface from vif id
        """
        return Main.kernel.vif_index_to_name_dic[interface_id]

    @staticmethod
    def get_interface(kernel_tree, interface_id):
        """
        Get PIM interface from interface id
        """
        interface_name = kernel_tree.get_interface_name(interface_id)
        return Main.interfaces.get(interface_name, None)

    @staticmethod
    def get_membership_interface(kernel_tree, interface_id):
        """
        Get IGMP interface from interface id
        """
        interface_name = kernel_tree.get_interface_name(interface_id)
        return Main.igmp_interfaces.get(interface_name, None)  # type: InterfaceIGMP

    @staticmethod
    def get_kernel():
        """
        Get kernel
        """
        return Main.kernel


class KernelEntry6Interface(KernelEntryInterface):
    @staticmethod
    def get_outbound_interfaces_indexes(kernel_tree):
        """
        Get OIL of this tree
        """
        outbound_indexes = [0] * 8
        for (index, state) in kernel_tree.interface_state.items():
            outbound_indexes[index // 32] |= state.is_forwarding() << (index % 32)
        return outbound_indexes

    @staticmethod
    def get_interface_name(interface_id):
        """
        Get name of interface from vif id
        """
        return Main.kernel_v6.vif_index_to_name_dic[interface_id]

    @staticmethod
    def get_interface(kernel_tree, interface_id):
        """
        Get PIM interface from interface id
        """
        interface_name = kernel_tree.get_interface_name(interface_id)
        return Main.interfaces_v6.get(interface_name, None)

    @staticmethod
    def get_membership_interface(kernel_tree, interface_id):
        """
        Get MLD interface from interface id
        """
        interface_name = kernel_tree.get_interface_name(interface_id)
        return Main.mld_interfaces.get(interface_name, None)  # type: InterfaceMLD

    @staticmethod
    def get_kernel():
        """
        Get kernel
        """
        return Main.kernel_v6



import ipaddress
import struct
import socket
'''
 0                   1                   2                   3
 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|  Addr Family  | Encoding Type |  Rsrvd  |S|W|R|  Mask Len     |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                         Source Address
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+...
'''
class PacketPimEncodedSourceAddress:
    PIM_ENCODED_SOURCE_ADDRESS_HDR = "! BBBB %s"
    PIM_ENCODED_SOURCE_ADDRESS_HDR_WITHOUT_SOURCE_ADDRESS = "! BBBB"


    IPV4_HDR = "4s"
    IPV6_HDR = "16s"

    # TODO ver melhor versao ip
    PIM_ENCODED_SOURCE_ADDRESS_HDR_WITHOUT_SOURCE_ADDRESS_LEN = struct.calcsize(PIM_ENCODED_SOURCE_ADDRESS_HDR_WITHOUT_SOURCE_ADDRESS)
    PIM_ENCODED_SOURCE_ADDRESS_HDR_LEN = struct.calcsize(PIM_ENCODED_SOURCE_ADDRESS_HDR % IPV4_HDR)
    PIM_ENCODED_SOURCE_ADDRESS_HDR_LEN_IPV6 = struct.calcsize(PIM_ENCODED_SOURCE_ADDRESS_HDR % IPV6_HDR)

    FAMILY_RESERVED = 0
    FAMILY_IPV4 = 1
    FAMILY_IPV6 = 2

    RESERVED_AND_SWR_BITS = 0

    def __init__(self, source_address, mask_len=None):
        if type(source_address) not in (str, bytes):
            raise Exception
        if type(source_address) is bytes:
            source_address = socket.inet_ntoa(source_address)
        self.source_address = source_address
        self.mask_len = mask_len

    def bytes(self) -> bytes:
        (string_ip_hdr, hdr_addr_family, socket_family) = PacketPimEncodedSourceAddress.get_ip_info(self.source_address)

        mask_len = self.mask_len
        if mask_len is None:
            mask_len = 8 * struct.calcsize(string_ip_hdr)
        ip = socket.inet_pton(socket_family, self.source_address)

        msg = struct.pack(PacketPimEncodedSourceAddress.PIM_ENCODED_SOURCE_ADDRESS_HDR % string_ip_hdr, hdr_addr_family, 0,
                          PacketPimEncodedSourceAddress.RESERVED_AND_SWR_BITS, mask_len, ip)
        return msg

    @staticmethod
    def get_ip_info(ip):
        version = ipaddress.ip_address(ip).version
        if version == 4:
            return (PacketPimEncodedSourceAddress.IPV4_HDR, PacketPimEncodedSourceAddress.FAMILY_IPV4, socket.AF_INET)
        elif version == 6:
            return (PacketPimEncodedSourceAddress.IPV6_HDR, PacketPimEncodedSourceAddress.FAMILY_IPV6, socket.AF_INET6)
        else:
            raise Exception("Unknown address family")

    def __len__(self):
        version = ipaddress.ip_address(self.source_address).version
        if version == 4:
            return self.PIM_ENCODED_SOURCE_ADDRESS_HDR_LEN
        elif version == 6:
            return self.PIM_ENCODED_SOURCE_ADDRESS_HDR_LEN_IPV6
        else:
            raise Exception("Unknown address family")

    @staticmethod
    def parse_bytes(data: bytes):
        data_without_source_addr = data[0:PacketPimEncodedSourceAddress.PIM_ENCODED_SOURCE_ADDRESS_HDR_WITHOUT_SOURCE_ADDRESS_LEN]
        (addr_family, encoding, _, mask_len) = struct.unpack(PacketPimEncodedSourceAddress.PIM_ENCODED_SOURCE_ADDRESS_HDR_WITHOUT_SOURCE_ADDRESS, data_without_source_addr)

        data_source_addr = data[PacketPimEncodedSourceAddress.PIM_ENCODED_SOURCE_ADDRESS_HDR_WITHOUT_SOURCE_ADDRESS_LEN:]
        if addr_family == PacketPimEncodedSourceAddress.FAMILY_IPV4:
            (ip,) = struct.unpack("! " + PacketPimEncodedSourceAddress.IPV4_HDR, data_source_addr[:4])
            ip = socket.inet_ntop(socket.AF_INET, ip)
        elif addr_family == PacketPimEncodedSourceAddress.FAMILY_IPV6:
            (ip,) = struct.unpack("! " + PacketPimEncodedSourceAddress.IPV6_HDR, data_source_addr[:16])
            ip = socket.inet_ntop(socket.AF_INET6, ip)
        else:
            raise Exception("Unknown address family")

        if encoding != 0:
            print("unknown encoding")
            raise Exception

        return PacketPimEncodedSourceAddress(ip, mask_len)



# Protocol files
DAEMON_PROCESS_FILE = '/tmp/Daemon-pim{}.pid'
DAEMON_SOCKET = '/tmp/pim_uds_socket{}'
DAEMON_LOG_FOLDER = '/var/log/pimdm/'
DAEMON_LOG_STDOUT_FILE = DAEMON_LOG_FOLDER + 'stdout{}'

# PIM-DM TIMER VARIABLES
ASSERT_TIME = 180
GRAFT_RETRY_PERIOD = 3
JP_OVERRIDE_INTERVAL = 3.0
OVERRIDE_INTERVAL = 2.5
PROPAGATION_DELAY = 0.5
REFRESH_INTERVAL = 60  # State Refresh Interval
SOURCE_LIFETIME = 210
T_LIMIT = 210

# PIM-DM VARIABLES
HELLO_HOLD_TIME_NO_TIMEOUT = 0xFFFF
HELLO_HOLD_TIME = 160
HELLO_HOLD_TIME_TIMEOUT = 0

ASSERT_CANCEL_METRIC = 0xFFFFFFFF

# MULTIPLE TABLES SUPPORT
# Define which unicast routing table to be used for RPF checks and to get route metric information
# Default unicast routing table is 254
UNICAST_TABLE_ID = 254
# Define which multicast routing table to be used for setting multicast trees
# Default multicast routing table is 0
MULTICAST_TABLE_ID = 0



import struct
from abc import ABCMeta, abstractstaticmethod
from .PacketPimHelloOptions import PacketPimHelloOptions, PacketPimHelloStateRefreshCapable, PacketPimHelloGenerationID, PacketPimHelloLANPruneDelay, PacketPimHelloHoldtime

'''
 0                   1                   2                   3
 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|          Option Type          |         Option Length         |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                         Option Value                          |
|                              ...                              |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                               .                               |
|                               .                               |
|                               .                               |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|          Option Type          |         Option Length         |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                         Option Value                          |
|                              ...                              |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
'''
class PacketPimHello:
    PIM_TYPE = 0
    PIM_HDR_OPTS = "! HH"
    PIM_HDR_OPTS_LEN = struct.calcsize(PIM_HDR_OPTS)

    PIM_MSG_TYPES_LENGTH = {1: 2,
                            2: 4,
                            20: 4,
                            21: 4,
                            }
    # todo: pensar melhor na implementacao state refresh capable option...

    def __init__(self):
        self.options = {}

    '''
    def add_option(self, option_type: int, option_value: int or float):
        option_value = int(option_value)
        # if option_value requires more bits than the bits available for that field: option value will have all field bits = 1
        if option_type in self.PIM_MSG_TYPES_LENGTH and self.PIM_MSG_TYPES_LENGTH[option_type] * 8 < option_value.bit_length():
            option_value = (1 << (self.PIM_MSG_TYPES_LENGTH[option_type] * 8)) - 1
        self.options[option_type] = option_value
    '''

    def add_option(self, option: 'PacketPimHelloOptions'):
        #if option_type in self.PIM_MSG_TYPES_LENGTH and self.PIM_MSG_TYPES_LENGTH[option_type] * 8 < option_value.bit_length():
        #    option_value = (1 << (self.PIM_MSG_TYPES_LENGTH[option_type] * 8)) - 1

        self.options[option.type] = option


    def get_options(self):
        return self.options

    '''
    def bytes(self) -> bytes:
        res = b''
        for (option_type, option_value) in self.options.items():
            option_length = PacketPimHello.PIM_MSG_TYPES_LENGTH[option_type]
            type_length_hdr = struct.pack(PacketPimHello.PIM_HDR_OPTS, option_type, option_length)
            res += type_length_hdr + struct.pack("! " + str(option_length) + "s", option_value.to_bytes(option_length, byteorder='big'))
        return res
    '''

    def bytes(self) -> bytes:
        res = b''
        for option in self.options.values():
            res += option.bytes()
        return res



    def __len__(self):
        return len(self.bytes())


    '''
    @staticmethod
    def parse_bytes(data: bytes):
        pim_payload = PacketPimHello()
        while data != b'':
            (option_type, option_length) = struct.unpack(PacketPimHello.PIM_HDR_OPTS,
                                                         data[:PacketPimHello.PIM_HDR_OPTS_LEN])
            print(option_type, option_length)
            data = data[PacketPimHello.PIM_HDR_OPTS_LEN:]
            print(data)
            (option_value,) = struct.unpack("! " + str(option_length) + "s", data[:option_length])
            option_value_number = int.from_bytes(option_value, byteorder='big')
            print("option value: ", option_value_number)
            
            #options_list.append({"OPTION TYPE": option_type,
            #                     "OPTION LENGTH": option_length,
            #                     "OPTION VALUE": option_value_number
            #                     })
            
            pim_payload.add_option(option_type, option_value_number)
            data = data[option_length:]

        return pim_payload
    '''

    @staticmethod
    def parse_bytes(data: bytes):
        pim_payload = PacketPimHello()
        while data != b'':
            option = PacketPimHelloOptions.parse_bytes(data)
            option_length = len(option)
            data = data[option_length:]
            pim_payload.add_option(option)
        return pim_payload



import socket
from abc import ABCMeta, abstractmethod
import threading
import traceback


class Interface(metaclass=ABCMeta):
    MCAST_GRP = '224.0.0.13'

    def __init__(self, interface_name, recv_socket, send_socket, vif_index):
        self.interface_name = interface_name

        # virtual interface index for the multicast routing table
        self.vif_index = vif_index

        # set receive socket and send socket
        self._send_socket = send_socket
        self._recv_socket = recv_socket
        self.interface_enabled = False

    def _enable(self):
        """
        Enable this interface
        This will start a thread to be executed in the background to be used in the reception of control packets
        """
        self.interface_enabled = True
        # run receive method in background
        receive_thread = threading.Thread(target=self.receive)
        receive_thread.daemon = True
        receive_thread.start()

    def receive(self):
        """
        Method that will be executed in the background for the reception of control packets
        """
        while self.interface_enabled:
            try:
                (raw_bytes, ancdata, _, src_addr) = self._recv_socket.recvmsg(256 * 1024, 500)
                if raw_bytes:
                    self._receive(raw_bytes, ancdata, src_addr)
            except Exception:
                traceback.print_exc()
                continue

    @abstractmethod
    def _receive(self, raw_bytes, ancdata, src_addr):
        """
        Subclass method to be implemented
        This method will be invoked whenever a new control packet is received
        """
        raise NotImplementedError

    def send(self, data: bytes, group_ip: str):
        """
        Send a control packet through this interface
        Explicitly destined to group_ip (can be unicast or multicast IP)
        """
        if self.interface_enabled and data:
            self._send_socket.sendto(data, (group_ip, 0))

    def remove(self):
        """
        This interface is no longer active....
        Clear all state regarding it
        """
        self.interface_enabled = False
        try:
            self._recv_socket.shutdown(socket.SHUT_RDWR)
        except Exception:
            pass
        self._recv_socket.close()
        self._send_socket.close()

    def is_enabled(self):
        """
        Verify if this interface is enabled
        """
        return self.interface_enabled

    @abstractmethod
    def get_ip(self):
        """
        Get IP of this interface
        """
        raise NotImplementedError



import struct
import socket


class PacketIpHeader:
    """
     0                   1                   2                   3
     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |Version|
    +-+-+-+-+
    """
    IP_HDR = "! B"
    IP_HDR_LEN = struct.calcsize(IP_HDR)

    def __init__(self, ver, hdr_len):
        self.version = ver
        self.hdr_length = hdr_len

    def __len__(self):
        return self.hdr_length

    @staticmethod
    def parse_bytes(data: bytes):
        (verhlen, ) = struct.unpack(PacketIpHeader.IP_HDR, data[:PacketIpHeader.IP_HDR_LEN])
        ver = (verhlen & 0xF0) >> 4
        print("ver:", ver)
        return PACKET_HEADER.get(ver).parse_bytes(data)


class PacketIpv4Header(PacketIpHeader):
    """
     0                   1                   2                   3
     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |Version|  IHL  |Type of Service|          Total Length         |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |         Identification        |Flags|      Fragment Offset    |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |  Time to Live |    Protocol   |         Header Checksum       |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                       Source Address                          |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                    Destination Address                        |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                    Options                    |    Padding    |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    """
    IP_HDR = "! BBH HH BBH 4s 4s"
    IP_HDR_LEN = struct.calcsize(IP_HDR)

    def __init__(self, ver, hdr_len, ttl, proto, ip_src, ip_dst):
        super().__init__(ver, hdr_len)
        self.ttl = ttl
        self.proto = proto
        self.ip_src = ip_src
        self.ip_dst = ip_dst

    def __len__(self):
        return self.hdr_length

    @staticmethod
    def parse_bytes(data: bytes):
        (verhlen, tos, iplen, ipid, frag, ttl, proto, cksum, src, dst) = \
            struct.unpack(PacketIpv4Header.IP_HDR, data[:PacketIpv4Header.IP_HDR_LEN])

        ver = (verhlen & 0xf0) >> 4
        hlen = (verhlen & 0x0f) * 4

        '''
        "VER": ver,
        "HLEN": hlen,
        "TOS": tos,
        "IPLEN": iplen,
        "IPID": ipid,
        "FRAG": frag,
        "TTL": ttl,
        "PROTO": proto,
        "CKSUM": cksum,
        "SRC": socket.inet_ntoa(src),
        "DST": socket.inet_ntoa(dst)
        '''
        src_ip = socket.inet_ntoa(src)
        dst_ip = socket.inet_ntoa(dst)
        return PacketIpv4Header(ver, hlen, ttl, proto, src_ip, dst_ip)


class PacketIpv6Header(PacketIpHeader):
    """
     0                   1                   2                   3
     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |Version| Traffic Class |           Flow Label                  |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |         Payload Length        |  Next Header  |   Hop Limit   |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                                                               |
    +                                                               +
    |                                                               |
    +                         Source Address                        +
    |                                                               |
    +                                                               +
    |                                                               |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                                                               |
    +                                                               +
    |                                                               |
    +                      Destination Address                      +
    |                                                               |
    +                                                               +
    |                                                               |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    """
    IP6_HDR = "! I HBB 16s 16s"
    IP6_HDR_LEN = struct.calcsize(IP6_HDR)

    def __init__(self, ver, next_header, hop_limit, ip_src, ip_dst):
        # TODO: confirm hdr_length in case of multiple options/headers
        super().__init__(ver, PacketIpv6Header.IP6_HDR_LEN)
        self.next_header = next_header
        self.hop_limit = hop_limit
        self.ip_src = ip_src
        self.ip_dst = ip_dst

    def __len__(self):
        return PacketIpv6Header.IP6_HDR_LEN

    @staticmethod
    def parse_bytes(data: bytes):
        (ver_tc_fl, _, next_header, hop_limit, src, dst) = \
            struct.unpack(PacketIpv6Header.IP6_HDR, data[:PacketIpv6Header.IP6_HDR_LEN])

        ver = (ver_tc_fl & 0xf0000000) >> 28
        #tc = (ver_tc_fl & 0x0ff00000) >> 20
        #fl = (ver_tc_fl & 0x000fffff)
        '''
        "VER": ver,
        "TRAFFIC CLASS": tc,
        "FLOW LABEL": fl,
        "PAYLOAD LEN": payload_length,
        "NEXT HEADER": next_header,
        "HOP LIMIT": hop_limit,
        "SRC": socket.inet_atop(socket.AF_INET6, src),
        "DST": socket.inet_atop(socket.AF_INET6, dst)
        '''

        src_ip = socket.inet_ntop(socket.AF_INET6, src)
        dst_ip = socket.inet_ntop(socket.AF_INET6, dst)
        return PacketIpv6Header(ver, next_header, hop_limit, src_ip, dst_ip)


PACKET_HEADER = {
    4: PacketIpv4Header,
    6: PacketIpv6Header,
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Read Write Lock
"""

import threading
import time

class RWLockRead(object):
	"""
		A Read/Write lock giving preference to Reader
	"""
	def __init__(self):
		self.V_ReadCount = 0
		self.A_Resource = threading.Lock()
		self.A_LockReadCount = threading.Lock()
	class _aReader(object):
		def __init__(self, p_RWLock):
			self.A_RWLock = p_RWLock
			self.V_Locked = False

		def acquire(self, blocking=1, timeout=-1):
			p_TimeOut = None if (blocking and timeout < 0) else (timeout if blocking else 0)
			c_DeadLine = None if p_TimeOut is None else (time.time() + p_TimeOut)
			if not self.A_RWLock.A_LockReadCount.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0, c_DeadLine - time.time())):
				return False
			self.A_RWLock.V_ReadCount += 1
			if self.A_RWLock.V_ReadCount == 1:
				if not self.A_RWLock.A_Resource.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0, c_DeadLine - time.time())):
					self.A_RWLock.V_ReadCount -= 1
					self.A_RWLock.A_LockReadCount.release()
					return False
			self.A_RWLock.A_LockReadCount.release()
			self.V_Locked = True
			return True

		def release(self):
			if not self.V_Locked: raise RuntimeError("cannot release un-acquired lock")
			self.V_Locked = False
			self.A_RWLock.A_LockReadCount.acquire()
			self.A_RWLock.V_ReadCount -= 1
			if self.A_RWLock.V_ReadCount == 0:
				self.A_RWLock.A_Resource.release()
			self.A_RWLock.A_LockReadCount.release()
		def locked(self):
			return self.V_Locked
		def __enter__(self):
			self.acquire()
		def __exit__(self, p_Type, p_Value, p_Traceback):
			self.release()
	class _aWriter(object):
		def __init__(self, p_RWLock):
			self.A_RWLock = p_RWLock
			self.V_Locked = False
		def acquire(self, blocking=1, timeout=-1):
			self.V_Locked = self.A_RWLock.A_Resource.acquire(blocking, timeout)
			return self.V_Locked
		def release(self):
			if not self.V_Locked: raise RuntimeError("cannot release un-acquired lock")
			self.V_Locked = False
			self.A_RWLock.A_Resource.release()
		def locked(self):
			return self.V_Locked
		def __enter__(self):
			self.acquire()
		def __exit__(self, p_Type, p_Value, p_Traceback):
			self.release()
	def genRlock(self):
		"""
			Generate a reader lock
		"""
		return RWLockRead._aReader(self)
	def genWlock(self):
		"""
			Generate a writer lock
		"""
		return RWLockRead._aWriter(self)

class RWLockWrite(object):
	"""
		A Read/Write lock giving preference to Writer
	"""
	def __init__(self):
		self.V_ReadCount = 0
		self.V_WriteCount = 0
		self.A_LockReadCount = threading.Lock()
		self.A_LockWriteCount = threading.Lock()
		self.A_LockReadEntry = threading.Lock()
		self.A_LockReadTry = threading.Lock()
		self.A_Resource = threading.Lock()
	class _aReader(object):
		def __init__(self, p_RWLock):
			self.A_RWLock = p_RWLock
			self.V_Locked = False
		def acquire(self, blocking=1, timeout=-1):
			p_TimeOut = None if (blocking and timeout < 0) else (timeout if blocking else 0)
			c_DeadLine = None if p_TimeOut is None else (time.time() + p_TimeOut)
			if not self.A_RWLock.A_LockReadEntry.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0, c_DeadLine - time.time())):
				return False
			if not self.A_RWLock.A_LockReadTry.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0, c_DeadLine - time.time())):
				self.A_RWLock.A_LockReadEntry.release()
				return False
			if not self.A_RWLock.A_LockReadCount.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0, c_DeadLine - time.time())):
				self.A_RWLock.A_LockReadTry.release()
				self.A_RWLock.A_LockReadEntry.release()
				return False
			self.A_RWLock.V_ReadCount += 1
			if (self.A_RWLock.V_ReadCount == 1):
				if not self.A_RWLock.A_Resource.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0, c_DeadLine - time.time())):
					self.A_RWLock.A_LockReadTry.release()
					self.A_RWLock.A_LockReadEntry.release()
					self.A_RWLock.V_ReadCount -= 1
					self.A_RWLock.A_LockReadCount.release()
					return False
			self.A_RWLock.A_LockReadCount.release()
			self.A_RWLock.A_LockReadTry.release()
			self.A_RWLock.A_LockReadEntry.release()
			self.V_Locked = True
			return True
		def release(self):
			if not self.V_Locked: raise RuntimeError("cannot release un-acquired lock")
			self.V_Locked = False
			self.A_RWLock.A_LockReadCount.acquire()
			self.A_RWLock.V_ReadCount -= 1
			if (self.A_RWLock.V_ReadCount == 0):
				self.A_RWLock.A_Resource.release()
			self.A_RWLock.A_LockReadCount.release()
		def locked(self):
			return self.V_Locked
		def __enter__(self):
			self.acquire()
		def __exit__(self, p_Type, p_Value, p_Traceback):
			self.release()
	class _aWriter(object):
		def __init__(self, p_RWLock):
			self.A_RWLock = p_RWLock
			self.V_Locked = False
		def acquire(self, blocking=1, timeout=-1):
			p_TimeOut = None if (blocking and timeout < 0) else (timeout if blocking else 0)
			c_DeadLine = None if p_TimeOut is None else (time.time() + p_TimeOut)
			if not self.A_RWLock.A_LockWriteCount.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0, c_DeadLine - time.time())):
				return False
			self.A_RWLock.V_WriteCount += 1
			if (self.A_RWLock.V_WriteCount == 1):
				if not self.A_RWLock.A_LockReadTry.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0, c_DeadLine - time.time())):
					self.A_RWLock.V_WriteCount -= 1
					self.A_RWLock.A_LockWriteCount.release()
					return False
			self.A_RWLock.A_LockWriteCount.release()
			if not self.A_RWLock.A_Resource.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0, c_DeadLine - time.time())):
				self.A_RWLock.A_LockWriteCount.acquire()
				self.A_RWLock.V_WriteCount -= 1
				if self.A_RWLock.V_WriteCount == 0:
					self.A_RWLock.A_LockReadTry.release()
				self.A_RWLock.A_LockWriteCount.release()
				return False
			self.V_Locked = True
			return True
		def release(self):
			if not self.V_Locked: raise RuntimeError("cannot release un-acquired lock")
			self.V_Locked = False
			self.A_RWLock.A_Resource.release()
			self.A_RWLock.A_LockWriteCount.acquire()
			self.A_RWLock.V_WriteCount -= 1
			if (self.A_RWLock.V_WriteCount == 0):
				self.A_RWLock.A_LockReadTry.release()
			self.A_RWLock.A_LockWriteCount.release()
		def locked(self):
			return self.V_Locked
		def __enter__(self):
			self.acquire()
		def __exit__(self, p_Type, p_Value, p_Traceback):
			self.release()
	def genRlock(self):
		"""
			Generate a reader lock
		"""
		return RWLockWrite._aReader(self)
	def genWlock(self):
		"""
			Generate a writer lock
		"""
		return RWLockWrite._aWriter(self)

class RWLockFair(object):
	"""
		A Read/Write lock giving fairness to both Reader and Writer
	"""
	def __init__(self):
		self.V_ReadCount = 0
		self.A_LockReadCount = threading.Lock()
		self.A_LockRead = threading.Lock()
		self.A_LockWrite = threading.Lock()
	class _aReader(object):
		def __init__(self, p_RWLock):
			self.A_RWLock = p_RWLock
			self.V_Locked = False
		def acquire(self, blocking=1, timeout=-1):
			p_TimeOut = None if (blocking and timeout < 0) else (timeout if blocking else 0)
			c_DeadLine = None if p_TimeOut is None else (time.time() + p_TimeOut)
			if not self.A_RWLock.A_LockRead.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0, c_DeadLine - time.time())):
				return False
			if not self.A_RWLock.A_LockReadCount.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0, c_DeadLine - time.time())):
				self.A_RWLock.A_LockRead.release()
				return False
			self.A_RWLock.V_ReadCount += 1
			if self.A_RWLock.V_ReadCount == 1:
				if not self.A_RWLock.A_LockWrite.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0, c_DeadLine - time.time())):
					self.A_RWLock.V_ReadCount -= 1
					self.A_RWLock.A_LockReadCount.release()
					self.A_RWLock.A_LockRead.release()
					return False
			self.A_RWLock.A_LockReadCount.release()
			self.A_RWLock.A_LockRead.release()
			self.V_Locked = True
			return True
		def release(self):
			if not self.V_Locked: raise RuntimeError("cannot release un-acquired lock")
			self.V_Locked = False
			self.A_RWLock.A_LockReadCount.acquire()
			self.A_RWLock.V_ReadCount -= 1
			if self.A_RWLock.V_ReadCount == 0:
				self.A_RWLock.A_LockWrite.release()
			self.A_RWLock.A_LockReadCount.release()
		def locked(self):
			return self.V_Locked
		def __enter__(self):
			self.acquire()
		def __exit__(self, p_Type, p_Value, p_Traceback):
			self.release()
	class _aWriter(object):
		def __init__(self, p_RWLock):
			self.A_RWLock = p_RWLock
			self.V_Locked = False
		def acquire(self, blocking=1, timeout=-1):
			p_TimeOut = None if (blocking and timeout < 0) else (timeout if blocking else 0)
			c_DeadLine = None if p_TimeOut is None else (time.time() + p_TimeOut)
			if not self.A_RWLock.A_LockRead.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0, c_DeadLine - time.time())):
				return False
			if not self.A_RWLock.A_LockWrite.acquire(blocking=1, timeout=-1 if c_DeadLine is None else max(0, c_DeadLine - time.time())):
				self.A_RWLock.A_LockRead.release()
				return False
			self.V_Locked = True
			return True
		def release(self):
			if not self.V_Locked: raise RuntimeError("cannot release un-acquired lock")
			self.V_Locked = False
			self.A_RWLock.A_LockWrite.release()
			self.A_RWLock.A_LockRead.release()
		def locked(self):
			return self.V_Locked
		def __enter__(self):
			self.acquire()
		def __exit__(self, p_Type, p_Value, p_Traceback):
			self.release()
	def genRlock(self):
		"""
			Generate a reader lock
		"""
		return RWLockFair._aReader(self)
	def genWlock(self):
		"""
			Generate a writer lock
		"""
		return RWLockFair._aWriter(self)
